import asyncio
import os
import threading
import traceback
import uuid
import webbrowser
from pathlib import Path
from typing import Any, Optional

import httpx
import typer
from arcadepy import Arcade
from arcadepy.types import AuthorizationResponse
from openai import OpenAI, OpenAIError
from rich.console import Console
from rich.markup import escape
from rich.text import Text
from tqdm import tqdm

import arcade_cli.worker as worker
from arcade_cli.authn import LocalAuthCallbackServer, check_existing_login
from arcade_cli.constants import (
    CREDENTIALS_FILE_PATH,
    LOCALHOST,
    PROD_CLOUD_HOST,
    PROD_ENGINE_HOST,
)
from arcade_cli.deployment import Deployment
from arcade_cli.display import (
    display_arcade_chat_header,
    display_eval_results,
    display_tool_messages,
)
from arcade_cli.show import show_logic
from arcade_cli.utils import (
    OrderCommands,
    compute_base_url,
    compute_login_url,
    get_eval_files,
    get_today_context,
    get_user_input,
    handle_chat_interaction,
    handle_tool_authorization,
    handle_user_command,
    is_authorization_pending,
    load_eval_suites,
    log_engine_health,
    require_dependency,
    validate_and_get_config,
    version_callback,
)

cli = typer.Typer(
    cls=OrderCommands,
    add_completion=False,
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=True,
    rich_markup_mode="markdown",
)


cli.add_typer(
    worker.app,
    name="worker",
    help="Manage deployments of tool servers (logs, list, etc)",
    rich_help_panel="Deployment",
)
console = Console()


def handle_cli_error(
    message: str, error: Exception | None = None, debug: bool = True, should_exit: bool = True
) -> None:
    """Handle CLI error reporting with optional debug traceback and exit."""
    if error and debug:
        console.print(f"❌ {message}: {traceback.format_exc()}", style="bold red")
    elif error:
        console.print(f"❌ {message}: {escape(str(error))}", style="bold red")
    else:
        console.print(f"❌ {message}", style="bold red")

    if should_exit:
        raise typer.Exit(code=1)


@cli.command(help="Log in to Arcade Cloud", rich_help_panel="User")
def login(
    host: str = typer.Option(
        PROD_CLOUD_HOST,
        "-h",
        "--host",
        help="The Arcade Cloud host to log in to.",
    ),
    port: Optional[int] = typer.Option(
        None,
        "-p",
        "--port",
        help="The port of the Arcade Cloud host (if running locally).",
    ),
    callback_host: str = typer.Option(
        None,
        "--callback-host",
        help="The host to use to complete the auth flow - this should be the same as the host that the CLI is running on. Include the port if needed.",
    ),
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
) -> None:
    """
    Logs the user into Arcade Cloud.
    """

    if check_existing_login():
        console.print("\nTo log out and delete your locally-stored credentials, use ", end="")
        console.print("arcade logout", style="bold green", end="")
        console.print(".\n")
        return

    # Start the HTTP server in a new thread
    state = str(uuid.uuid4())
    auth_server = LocalAuthCallbackServer(state)
    server_thread = threading.Thread(target=auth_server.run_server)
    server_thread.start()

    try:
        # Open the browser for user login
        login_url = compute_login_url(host, state, port, callback_host)

        console.print("Opening a browser to log you in...")
        if not webbrowser.open(login_url):
            console.print(
                f"If a browser doesn't open automatically, copy this URL and paste it into your browser: {login_url}",
                style="dim",
            )

        # Wait for the server thread to finish
        server_thread.join()
    except KeyboardInterrupt:
        auth_server.shutdown_server()
    except Exception as e:
        handle_cli_error("Login failed", e, debug)
    finally:
        if server_thread.is_alive():
            server_thread.join()  # Ensure the server thread completes and cleans up


@cli.command(help="Log out of Arcade Cloud", rich_help_panel="User")
def logout(
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
) -> None:
    """
    Logs the user out of Arcade Cloud.
    """
    try:
        # If the credentials file exists, delete it
        if os.path.exists(CREDENTIALS_FILE_PATH):
            os.remove(CREDENTIALS_FILE_PATH)
            console.print("You're now logged out.", style="bold")
        else:
            console.print("You're not logged in.", style="bold red")
    except Exception as e:
        handle_cli_error("Logout failed", e, debug)


@cli.command(
    help="Create a new toolkit package directory. Example usage: arcade new my_toolkit",
    rich_help_panel="Tool Development",
)
def new(
    toolkit_name: str = typer.Argument(
        help="The name of the toolkit to create",
        metavar="TOOLKIT_NAME",
    ),
    directory: str = typer.Option(os.getcwd(), "--dir", help="tools directory path"),
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
) -> None:
    """
    Creates a new toolkit with the given name, description, and result type.
    """
    from arcade_cli.new import create_new_toolkit

    try:
        create_new_toolkit(directory, toolkit_name)
    except Exception as e:
        handle_cli_error("Failed to create new Toolkit", e, debug)


@cli.command(
    help="Show the installed toolkits or details of a specific tool",
    rich_help_panel="Tool Development",
)
def show(
    toolkit: Optional[str] = typer.Option(
        None, "-T", "--toolkit", help="The toolkit to show the tools of"
    ),
    tool: Optional[str] = typer.Option(
        None, "-t", "--tool", help="The specific tool to show details for"
    ),
    host: str = typer.Option(
        PROD_ENGINE_HOST,
        "-h",
        "--host",
        help="The Arcade Engine address to show the tools/toolkits of.",
    ),
    local: bool = typer.Option(
        False,
        "--local",
        "-l",
        help="Show the local environment's catalog instead of an Arcade Engine's catalog.",
    ),
    port: Optional[int] = typer.Option(
        None,
        "-p",
        "--port",
        help="The port of the Arcade Engine.",
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine. If not specified, the connection will use TLS if the engine URL uses a 'https' scheme.",
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
    ),
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
) -> None:
    """
    Show the available toolkits or detailed information about a specific tool.
    """
    show_logic(toolkit, tool, host, local, port, force_tls, force_no_tls, debug)


@cli.command(
    help="Start a chat with a model in the terminal to test tools",
    rich_help_panel="Tool Development",
)
def chat(
    model: str = typer.Option("gpt-4o", "-m", "--model", help="The model to use for prediction."),
    stream: bool = typer.Option(
        False, "-s", "--stream", is_flag=True, help="Stream the tool output."
    ),
    prompt: str = typer.Option(None, "--prompt", help="The system prompt to use for the chat."),
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
    host: str = typer.Option(
        PROD_ENGINE_HOST,
        "-h",
        "--host",
        help="The Arcade Engine address to send chat requests to.",
    ),
    port: int = typer.Option(
        None,
        "-p",
        "--port",
        help="The port of the Arcade Engine.",
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine. If not specified, the connection will use TLS if the engine URL uses a 'https' scheme.",
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
    ),
) -> None:
    """
    Chat with a language model.
    """
    try:
        import readline
    except ImportError:
        console.print(
            "Readline is not available on this platform. Command history will be limited.",
            style="dim",
        )

    config = validate_and_get_config()
    base_url = compute_base_url(force_tls, force_no_tls, host, port)

    client = Arcade(api_key=config.api.key, base_url=base_url)
    user_email = config.user.email if config.user else None

    try:
        # start messages conversation
        history: list[dict[str, Any]] = []

        # Ground the LLM with today's date and day of the week to help when calling date-related tools
        # in case the user refers to relative dates (e.g. next Monday, last month, etc)
        today_context = get_today_context()

        if prompt:
            prompt = f"{today_context} {prompt}"
        else:
            prompt = today_context

            history.append({"role": "system", "content": prompt})

        display_arcade_chat_header(base_url, stream)

        # Try to hit /health endpoint on engine and warn if it is down
        log_engine_health(client)

        while True:
            console.print(
                f"\n[magenta][bold]User[/bold] {user_email}: [/magenta]"
                + "([bold][default]/?[/default][/bold] for help)"
            )

            user_input = get_user_input()

            # Add the input to history
            readline.add_history(user_input)

            if handle_user_command(
                user_input, history, host, port, force_tls, force_no_tls, show_logic
            ):
                continue

            history.append({"role": "user", "content": user_input})

            try:
                # TODO fixup configuration to remove this + "/v1" workaround
                openai_client = OpenAI(api_key=config.api.key, base_url=base_url + "/v1")
                chat_result = handle_chat_interaction(
                    openai_client, model, history, user_email, stream
                )

                history = chat_result.history
                tool_messages = chat_result.tool_messages
                tool_authorization = chat_result.tool_authorization

                # wait for tool authorizations to complete, if any
                if tool_authorization and is_authorization_pending(tool_authorization):
                    chat_result = handle_tool_authorization(
                        client,
                        AuthorizationResponse.model_validate(tool_authorization),
                        history,
                        openai_client,
                        model,
                        user_email,
                        stream,
                    )
                    history = chat_result.history
                    tool_messages = chat_result.tool_messages

            except OpenAIError as e:
                handle_cli_error("Arcade Chat failed", e, debug, should_exit=False)
                continue
            if debug:
                display_tool_messages(tool_messages)

    except KeyboardInterrupt:
        console.print("Chat stopped by user.", style="bold blue")
        typer.Exit()

    except RuntimeError as e:
        handle_cli_error("Failed to run tool", e, debug)


@cli.command(help="Run tool calling evaluations", rich_help_panel="Tool Development")
def evals(
    directory: str = typer.Argument(".", help="Directory containing evaluation files"),
    show_details: bool = typer.Option(False, "--details", "-d", help="Show detailed results"),
    max_concurrent: int = typer.Option(
        1,
        "--max-concurrent",
        "-c",
        help="Maximum number of concurrent evaluations (default: 1)",
    ),
    models: str = typer.Option(
        "gpt-4o",
        "--models",
        "-m",
        help="The models to use for evaluation (default: gpt-4o). Use commas to separate multiple models.",
    ),
    host: str = typer.Option(
        LOCALHOST,
        "-h",
        "--host",
        help="The Arcade Engine address to send chat requests to.",
    ),
    cloud: bool = typer.Option(
        False,
        "--cloud",
        help="Whether to run evaluations against the Arcade Cloud Engine. Overrides the 'host' option.",
    ),
    port: int = typer.Option(
        None,
        "-p",
        "--port",
        help="The port of the Arcade Engine.",
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine. If not specified, the connection will use TLS if the engine URL uses a 'https' scheme.",
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
    ),
    debug: bool = typer.Option(False, "--debug", help="Show debug information"),
) -> None:
    """
    Find all files starting with 'eval_' in the given directory,
    execute any functions decorated with @tool_eval, and display the results.
    """
    require_dependency(
        package_name="arcade_evals",
        command_name="evals",
        install_command=r"pip install 'arcade-ai\[evals]'",
    )
    # Although Evals does not depend on the TDK, some evaluations import the
    # ToolCatalog class from the TDK instead of from arcade_core, so we require
    # the TDK to run the evals CLI command to avoid possible import errors.
    require_dependency(
        package_name="arcade_tdk",
        command_name="evals",
        install_command=r"pip install arcade-tdk",
    )

    config = validate_and_get_config()

    host = PROD_ENGINE_HOST if cloud else host
    base_url = compute_base_url(force_tls, force_no_tls, host, port)

    models_list = models.split(",")  # Use 'models_list' to avoid shadowing

    eval_files = get_eval_files(directory)
    if not eval_files:
        return

    console.print(
        Text.assemble(
            ("\nRunning evaluations against Arcade Engine at ", "bold"),
            (base_url, "bold blue"),
        )
    )

    # Try to hit /health endpoint on engine and warn if it is down
    with Arcade(api_key=config.api.key, base_url=base_url) as client:
        log_engine_health(client)

    # Use the new function to load eval suites
    eval_suites = load_eval_suites(eval_files)

    if not eval_suites:
        console.print("No evaluation suites to run.", style="bold yellow")
        return

    if show_details:
        suite_label = "suite" if len(eval_suites) == 1 else "suites"
        console.print(
            f"\nFound {len(eval_suites)} {suite_label} in the evaluation files.",
            style="bold",
        )

    async def run_evaluations() -> None:
        all_evaluations = []
        tasks = []
        for suite_func in eval_suites:
            console.print(
                Text.assemble(
                    ("Running evaluations in ", "bold"),
                    (suite_func.__name__, "bold blue"),
                )
            )
            for model in models_list:
                task = asyncio.create_task(
                    suite_func(
                        config=config,
                        base_url=base_url,
                        model=model,
                        max_concurrency=max_concurrent,
                    )
                )
                tasks.append(task)

        # Track progress and results as suite functions complete
        with tqdm(total=len(tasks), desc="Evaluations Progress") as pbar:
            results = []
            for f in asyncio.as_completed(tasks):
                results.append(await f)
                pbar.update(1)

        # TODO error handling on each eval
        all_evaluations.extend(results)
        display_eval_results(all_evaluations, show_details=show_details)

    try:
        asyncio.run(run_evaluations())
    except Exception as e:
        handle_cli_error("Failed to run evaluations", e, debug)


@cli.command(
    help="Start tool server worker with locally installed tools",
    rich_help_panel="Launch",
)
def serve(
    host: str = typer.Option(
        "127.0.0.1",
        help="Host for the app, from settings by default.",
        show_default=True,
    ),
    port: int = typer.Option(
        "8002", "-p", "--port", help="Port for the app, defaults to ", show_default=True
    ),
    disable_auth: bool = typer.Option(
        False,
        "--no-auth",
        help="Disable authentication for the worker. Not recommended for production.",
        show_default=True,
    ),
    otel_enable: bool = typer.Option(
        False, "--otel-enable", help="Send logs to OpenTelemetry", show_default=True
    ),
    mcp: bool = typer.Option(
        False, "--mcp", help="Run as a local MCP server over stdio", show_default=True
    ),
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
    reload: bool = typer.Option(
        False,
        "--reload",
        help="Enable auto-reloading when toolkit or server files change.",
        show_default=True,
    ),
) -> None:
    """
    Start a local Arcade Worker server.
    """
    require_dependency(
        package_name="arcade_serve",
        command_name="serve",
        install_command=r"pip install 'arcade-serve'",
    )

    from arcade_cli.serve import serve_default_worker

    try:
        serve_default_worker(
            host,
            port,
            disable_auth=disable_auth,
            enable_otel=otel_enable,
            debug=debug,
            mcp=mcp,
            reload=reload,
        )
    except KeyboardInterrupt:
        typer.Exit()
    except Exception as e:
        handle_cli_error("Failed to start Arcade Worker", e, debug)


@cli.command(
    help="Start a server with locally installed Arcade tools", rich_help_panel="Launch", hidden=True
)
def workerup(
    host: str = typer.Option(
        "127.0.0.1",
        help="Host for the app, from settings by default.",
        show_default=True,
    ),
    port: int = typer.Option(
        "8002", "-p", "--port", help="Port for the app, defaults to ", show_default=True
    ),
    disable_auth: bool = typer.Option(
        False,
        "--no-auth",
        help="Disable authentication for the worker. Not recommended for production.",
        show_default=True,
    ),
    otel_enable: bool = typer.Option(
        False, "--otel-enable", help="Send logs to OpenTelemetry", show_default=True
    ),
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
) -> None:
    """
    Starts the worker with host, port, and reload options. Uses
    Uvicorn as ASGI worker. Parameters allow runtime configuration.
    """
    require_dependency(
        package_name="arcade_serve",
        command_name="worker",
        install_command=r"pip install 'arcade-serve'",
    )

    from arcade_cli.serve import serve_default_worker

    try:
        serve_default_worker(
            host,
            port,
            disable_auth=disable_auth,
            enable_otel=otel_enable,
            debug=debug,
        )
    except KeyboardInterrupt:
        typer.Exit()
    except Exception as e:
        handle_cli_error("Failed to start Arcade Toolkit Server", e, debug)


@cli.command(help="Deploy toolkits to Arcade Cloud", rich_help_panel="Deployment")
def deploy(
    deployment_file: str = typer.Option(
        "worker.toml", "--deployment-file", "-d", help="The deployment file to deploy."
    ),
    cloud_host: str = typer.Option(
        PROD_CLOUD_HOST,
        "--cloud-host",
        "-c",
        help="The Arcade Cloud host to deploy to.",
        hidden=True,
    ),
    cloud_port: int = typer.Option(
        None,
        "--cloud-port",
        "-cp",
        help="The port of the Arcade Cloud host.",
        hidden=True,
    ),
    host: str = typer.Option(
        PROD_ENGINE_HOST,
        "--host",
        "-h",
        help="The Arcade Engine host to register the worker to.",
    ),
    port: int = typer.Option(
        None,
        "--port",
        "-p",
        help="The port of the Arcade Engine host.",
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine. If not specified, the connection will use TLS if the engine URL uses a 'https' scheme.",
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
    ),
    debug: bool = typer.Option(False, "--debug", help="Show debug information"),
) -> None:
    """
    Deploy a worker to Arcade Cloud.
    """

    config = validate_and_get_config()
    engine_url = compute_base_url(force_tls, force_no_tls, host, port)
    engine_client = Arcade(api_key=config.api.key, base_url=engine_url)
    cloud_url = compute_base_url(force_tls, force_no_tls, cloud_host, cloud_port)
    cloud_client = httpx.Client(
        base_url=cloud_url, headers={"Authorization": f"Bearer {config.api.key}"}
    )

    # Fetch deployment configuration
    try:
        deployment = Deployment.from_toml(Path(deployment_file))
    except Exception as e:
        handle_cli_error("Failed to parse deployment file", e, debug)

    with console.status(f"Deploying {len(deployment.worker)} workers"):
        for worker in deployment.worker:
            console.log(f"Deploying '{worker.config.id}...'", style="dim")
            try:
                # Attempt to deploy worker
                worker.request().execute(cloud_client, engine_client)
                console.log(f"✅ Worker '{worker.config.id}' deployed successfully.", style="dim")
            except Exception as e:
                handle_cli_error(f"Failed to deploy worker '{worker.config.id}'", e, debug)


@cli.command(help="Open the Arcade Dashboard in a web browser", rich_help_panel="User")
def dashboard(
    host: str = typer.Option(
        PROD_ENGINE_HOST,
        "-h",
        "--host",
        help="The Arcade Engine host that serves the dashboard.",
    ),
    port: Optional[int] = typer.Option(
        None,
        "-p",
        "--port",
        help="The port of the Arcade Engine.",
    ),
    local: bool = typer.Option(
        False,
        "--local",
        "-l",
        help="Open the local dashboard instead of the default remote dashboard.",
    ),
    force_tls: bool = typer.Option(
        False,
        "--tls",
        help="Whether to force TLS for the connection to the Arcade Engine.",
    ),
    force_no_tls: bool = typer.Option(
        False,
        "--no-tls",
        help="Whether to disable TLS for the connection to the Arcade Engine.",
    ),
    debug: bool = typer.Option(False, "--debug", "-d", help="Show debug information"),
) -> None:
    """Opens the Arcade Dashboard in a web browser.

    The Dashboard is a web-based Arcade user interface that is served by the Arcade Engine.
    """
    try:
        if local:
            host = "localhost"

        # Construct base URL (for both health check and dashboard)
        base_url = compute_base_url(force_tls, force_no_tls, host, port)
        dashboard_url = f"{base_url}/dashboard"

        # Try to hit /health endpoint on engine and warn if it is down
        config = validate_and_get_config()
        with Arcade(api_key=config.api.key, base_url=base_url) as client:
            log_engine_health(client)

        # Open the dashboard in a browser
        console.print(f"Opening Arcade Dashboard at {dashboard_url}")
        if not webbrowser.open(dashboard_url):
            console.print(
                f"If a browser doesn't open automatically, copy this URL and paste it into your browser: {dashboard_url}",
                style="dim",
            )
    except Exception as e:
        handle_cli_error("Failed to open dashboard", e, debug)


@cli.callback()
def main_callback(
    ctx: typer.Context,
    _: Optional[bool] = typer.Option(
        None,
        "-v",
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print version and exit.",
    ),
) -> None:
    excluded_commands = {
        login.__name__,
        logout.__name__,
        serve.__name__,
        workerup.__name__,
        dashboard.__name__,
    }
    if ctx.invoked_subcommand in excluded_commands:
        return

    if not check_existing_login(suppress_message=True):
        console.print("Not logged in to Arcade CLI. Use ", style="bold red", end="")
        console.print("arcade login", style="bold green")
        raise typer.Exit()



import json
from typing import Optional

from pydantic import BaseModel, field_validator, model_validator

from arcade_google_sheets.enums import CellErrorType, NumberFormatType
from arcade_google_sheets.types import CellValue


class CellErrorValue(BaseModel):
    """An error in a cell

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#ErrorValue
    """

    type: CellErrorType
    message: str


class CellExtendedValue(BaseModel):
    """The kinds of value that a cell in a spreadsheet can have

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#ExtendedValue
    """

    numberValue: float | None = None
    stringValue: str | None = None
    boolValue: bool | None = None
    formulaValue: str | None = None
    errorValue: Optional["CellErrorValue"] = None

    @model_validator(mode="after")
    def check_exactly_one_value(cls, instance):  # type: ignore[no-untyped-def]
        provided = [v for v in instance.__dict__.values() if v is not None]
        if len(provided) != 1:
            raise ValueError(
                "Exactly one of numberValue, stringValue, boolValue, "
                "formulaValue, or errorValue must be set."
            )
        return instance


class NumberFormat(BaseModel):
    """The format of a number

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#NumberFormat
    """

    pattern: str
    type: NumberFormatType


class CellFormat(BaseModel):
    """The format of a cell

    Partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellFormat
    """

    numberFormat: NumberFormat


class CellData(BaseModel):
    """Data about a specific cell

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellData
    """

    userEnteredValue: CellExtendedValue
    userEnteredFormat: CellFormat | None = None


class RowData(BaseModel):
    """Data about each cellin a row

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#RowData
    """

    values: list[CellData]


class GridData(BaseModel):
    """Data in the grid

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#GridData
    """

    startRow: int
    startColumn: int
    rowData: list[RowData]


class GridProperties(BaseModel):
    """Properties of a grid

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#GridProperties
    """

    rowCount: int
    columnCount: int


class SheetProperties(BaseModel):
    """Properties of a Sheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#SheetProperties
    """

    sheetId: int
    title: str
    gridProperties: GridProperties | None = None


class Sheet(BaseModel):
    """A Sheet in a spreadsheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#Sheet
    """

    properties: SheetProperties
    data: list[GridData] | None = None


class SpreadsheetProperties(BaseModel):
    """Properties of a spreadsheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#SpreadsheetProperties
    """

    title: str


class Spreadsheet(BaseModel):
    """A spreadsheet

    A partial implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets
    """

    properties: SpreadsheetProperties
    sheets: list[Sheet]


class SheetDataInput(BaseModel):
    """
    SheetDataInput models the cell data of a spreadsheet in a custom format.

    It is a dictionary mapping row numbers (as ints) to dictionaries that map
    column letters (as uppercase strings) to cell values (int, float, str, or bool).

    This model enforces that:
      - The outer keys are convertible to int.
      - The inner keys are alphabetic strings (normalized to uppercase).
      - All cell values are only of type int, float, str, or bool.

    The model automatically serializes (via `json_data()`)
    and validates the inner types.
    """

    data: dict[int, dict[str, CellValue]]

    @classmethod
    def _parse_json_if_string(cls, value):  # type: ignore[no-untyped-def]
        """Parses the value if it is a JSON string, otherwise returns it.

        Helper method for when validating the `data` field.
        """
        if isinstance(value, str):
            try:
                return json.loads(value)
            except json.JSONDecodeError as e:
                raise TypeError(f"Invalid JSON: {e}")
        return value

    @classmethod
    def _validate_row_key(cls, row_key) -> int:  # type: ignore[no-untyped-def]
        """Converts the row key to an integer, raising an error if conversion fails.

        Helper method for when validating the `data` field.
        """
        try:
            return int(row_key)
        except (ValueError, TypeError):
            raise TypeError(f"Row key '{row_key}' is not convertible to int.")

    @classmethod
    def _validate_inner_cells(cls, cells, row_int: int) -> dict:  # type: ignore[no-untyped-def]
        """Validates that 'cells' is a dict mapping column letters to valid cell values
        and normalizes the keys.

        Helper method for when validating the `data` field.
        """
        if not isinstance(cells, dict):
            raise TypeError(
                f"Value for row '{row_int}' must be a dict mapping column letters to cell values."
            )
        new_inner = {}
        for col_key, cell_value in cells.items():
            if not isinstance(col_key, str):
                raise TypeError(f"Column key '{col_key}' must be a string.")
            col_string = col_key.upper()
            if not col_string.isalpha():
                raise TypeError(f"Column key '{col_key}' is invalid. Must be alphabetic.")
            if not isinstance(cell_value, int | float | str | bool):
                raise TypeError(
                    f"Cell value for {col_string}{row_int} must be an int, float, str, or bool."
                )
            new_inner[col_string] = cell_value
        return new_inner

    @field_validator("data", mode="before")
    @classmethod
    def validate_and_convert_keys(cls, value):  # type: ignore[no-untyped-def]
        """
        Validates data when SheetDataInput is instantiated and converts it to the correct format.
        Uses private helper methods to parse JSON, validate row keys, and validate inner cell data.
        """
        if value is None:
            return {}

        value = cls._parse_json_if_string(value)
        if isinstance(value, dict):
            new_value = {}
            for row_key, cells in value.items():
                row_int = cls._validate_row_key(row_key)
                inner_cells = cls._validate_inner_cells(cells, row_int)
                new_value[row_int] = inner_cells
            return new_value

        raise TypeError("data must be a dict or a valid JSON string representing a dict")

    def json_data(self) -> str:
        """
        Serialize the sheet data to a JSON string.
        """
        return json.dumps(self.data)

    @classmethod
    def from_json(cls, json_str: str) -> "SheetDataInput":
        """
        Create a SheetData instance from a JSON string.
        """
        return cls.model_validate_json(json_str)



optional_file_picker_instructions_template = (
    "Ensure the user knows that they have the option to select and grant access permissions to "
    "additional files and folders via the Google Drive File Picker. "
    "The user can pick additional files and folders via the following link: {url}"
)



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

from arcade_outlook_mail import (
    create_and_send_email,
    reply_to_email,
    send_draft_email,
)
from arcade_outlook_mail.enums import ReplyType
from evals.additional_messages import (
    list_emails_with_pagination_token_additional_messages,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_tool(create_and_send_email, "OutlookMail")
catalog.add_tool(send_draft_email, "OutlookMail")
catalog.add_tool(reply_to_email, "OutlookMail")


@tool_eval()
def outlook_mail_send_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Outlook Mail tools."""
    suite = EvalSuite(
        name="Outlook Mail Send Evaluation",
        system_message=("You are an AI that has access to tools to send, read, and write emails."),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Create draft email",
        user_message=(
            "send an email to j@arcade.dev and e@arcade.dev. Title it 'Hello friends' and have it "
            "say 'I've gathered you all here to celebrate the launch of the new Arcade platform.'"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_and_send_email,
                args={
                    "subject": "Hello friends",
                    "body": "I've gathered you all here to celebrate the launch of the new Arcade platform.",
                    "to_recipients": ["j@arcade.dev", "e@arcade.dev"],
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=0.3),
            SimilarityCritic(critic_field="body", weight=0.3),
            BinaryCritic(critic_field="to_recipients", weight=0.4),
        ],
    )

    suite.add_case(
        name="Update draft email",
        user_message=(
            "forward the draft AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDwAAAFuxokOLZRtDncM4_x_WeUwAAAAC-dpvAAAA "
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_draft_email,
                args={
                    "message_id": "AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDwAAAFuxokOLZRtDncM4_x_WeUwAAAAC-dpvAAAA",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="message_id", weight=1),
        ],
    )

    suite.add_case(
        name="Reply all to email",
        user_message=("Reply to everyone - 'sounds good to me'"),
        expected_tool_calls=[
            ExpectedToolCall(
                func=reply_to_email,
                args={
                    "message_id": "AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDAAAAFuxokOLZRtDncM4_x_WeUwAAAABc_ezAAAA",
                    "body": "sounds good to me",
                    "reply_type": ReplyType.REPLY_ALL,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="message_id", weight=1 / 3),
            SimilarityCritic(critic_field="body", weight=1 / 3),
            BinaryCritic(critic_field="reply_type", weight=1 / 3),
        ],
        additional_messages=list_emails_with_pagination_token_additional_messages,
    )

    suite.add_case(
        name="Reply to email",
        user_message=("Reply to the account security team - 'sounds good to me'"),
        expected_tool_calls=[
            ExpectedToolCall(
                func=reply_to_email,
                args={
                    "message_id": "AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDAAAAFuxokOLZRtDncM4_x_WeUwAAAABc_ezAAAA",
                    "body": "sounds good to me",
                    "reply_type": ReplyType.REPLY,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="message_id", weight=1 / 3),
            SimilarityCritic(critic_field="body", weight=1 / 3),
            BinaryCritic(critic_field="reply_type", weight=1 / 3),
        ],
        additional_messages=list_emails_with_pagination_token_additional_messages,
    )

    return suite



import asyncio
import logging
import queue
import signal
import sys
import threading
from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    pass

from arcade_serve.mcp.server import MCPServer

logger = logging.getLogger("arcade.mcp")

T = TypeVar("T")


def stdio_reader(stdin: object, q: queue.Queue[str | None]) -> None:
    """Read lines from stdin and put them into a queue."""
    for line in stdin:  # type: ignore[attr-defined]
        q.put(line)
    q.put(None)


def stdio_writer(stdout: object, q: queue.Queue[str | None]) -> None:
    """Write messages from a queue to stdout."""
    try:
        while True:
            msg = q.get()
            if msg is None:
                break

            # Ensure message ends with a newline for proper JSON-RPC-over-stdio
            if not msg.endswith("\n"):
                msg += "\n"

            stdout.write(msg)  # type: ignore[attr-defined]
            stdout.flush()  # type: ignore[attr-defined]
    except Exception:
        logger.exception("Error in stdio writer")


class StdioServer(MCPServer):
    """
    Stdio server that handles signals and cleanup.
    """

    def __init__(
        self,
        tool_catalog: Any,
        enable_logging: bool = True,
        **client_kwargs: dict[str, Any],
    ):
        # Set up stdio-specific middleware configuration
        middleware_config = client_kwargs.get("middleware_config", {})
        middleware_config["stdio_mode"] = True
        client_kwargs["middleware_config"] = middleware_config

        super().__init__(tool_catalog, enable_logging, **client_kwargs)
        self.read_q: queue.Queue[str | None] = queue.Queue()
        self.write_q: queue.Queue[str | None] = queue.Queue()
        self.reader_thread: threading.Thread | None = None
        self.writer_thread: threading.Thread | None = None
        self.running = False
        self.shutdown_event = asyncio.Event()

    def start_io_threads(self) -> None:
        """Start stdio reader and writer threads."""
        self.reader_thread = threading.Thread(
            target=self._stdio_reader, args=(sys.stdin, self.read_q), daemon=True
        )
        self.writer_thread = threading.Thread(
            target=self._stdio_writer, args=(sys.stdout, self.write_q), daemon=True
        )
        self.reader_thread.start()
        self.writer_thread.start()

    def _stdio_reader(self, stdin: object, q: queue.Queue[str | None]) -> None:
        """Read lines from stdin and put them into a queue."""
        try:
            for line in stdin:  # type: ignore[attr-defined]
                if not self.running:
                    break
                q.put(line)
        except Exception:
            logger.exception("Error in stdio reader")
        finally:
            q.put(None)  # Signal EOF

    def _stdio_writer(self, stdout: object, q: queue.Queue[str | None]) -> None:
        """Write messages from a queue to stdout."""
        try:
            while self.running:
                msg = q.get()
                if msg is None:
                    break
                stdout.write(msg)  # type: ignore[attr-defined]
                stdout.flush()  # type: ignore[attr-defined]
        except Exception:
            logger.exception("Error in stdio writer")

    async def _read_stream(self) -> AsyncGenerator[str, None]:
        """Async generator that yields lines from the read queue."""
        while self.running:
            try:
                line = await asyncio.to_thread(self.read_q.get)
                if line is None:
                    break
                yield line
            except asyncio.CancelledError:
                break
            except Exception:
                logger.exception("Error reading from stdin")
                break

    async def shutdown(self) -> None:
        """Gracefully shut down the server."""
        if not self.running:
            return

        logger.info("Shutting down stdio server...")
        self.running = False

        # Signal shutdown to MCP server
        await self.shutdown()

        # Clean up IO queues and threads
        try:
            if self.read_q:
                self.read_q.put(None)
            if self.write_q:
                self.write_q.put(None)
        except Exception:
            logger.exception("Error during shutdown")

        # Signal completion
        self.shutdown_event.set()
        logger.info("Stdio server shutdown complete")

    async def run(self) -> None:
        """Run the stdio server with signal handling."""
        self.running = True

        # Set up signal handlers
        loop = asyncio.get_running_loop()
        for sig in (signal.SIGINT, signal.SIGTERM):
            try:
                loop.add_signal_handler(sig, lambda: asyncio.create_task(self.shutdown()))
            except NotImplementedError:
                # Windows doesn't support POSIX signals
                if sys.platform == "win32":
                    logger.warning("Signal handling not fully supported on Windows")
                else:
                    logger.warning(f"Failed to set up signal handler for {sig}")

        # Start IO threads
        self.start_io_threads()

        logger.info("Starting MCP server with stdio transport")

        # Create WriteStream class for MCP server
        class WriteStream:
            async def send(self_, message: str) -> None:
                if self.running:
                    await asyncio.to_thread(self.write_q.put, message)

        try:
            # Run MCP server connection
            await self.run_connection(self._read_stream(), WriteStream(), None)
        except asyncio.CancelledError:
            # Handle cancellation
            logger.info("Server operation cancelled")
        except KeyboardInterrupt:
            # Handle keyboard interrupt
            logger.info("Keyboard interrupt received")
        except Exception:
            # Handle unexpected errors
            logger.exception("Unexpected error")
        finally:
            # Ensure we clean up
            await self.shutdown()
            # Wait for shutdown to complete
            await self.shutdown_event.wait()



import httpx
from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError

from arcade_linkedin.tools.constants import LINKEDIN_BASE_URL


async def _send_linkedin_request(
    context: ToolContext,
    method: str,
    endpoint: str,
    params: dict | None = None,
    json_data: dict | None = None,
) -> httpx.Response:
    """
    Send an asynchronous request to the LinkedIn API.

    Args:
        context: The tool context containing the authorization token.
        method: The HTTP method (GET, POST, PUT, DELETE, etc.).
        endpoint: The API endpoint path (e.g., "/ugcPosts").
        params: Query parameters to include in the request.
        json_data: JSON data to include in the request body.

    Returns:
        The response object from the API request.

    Raises:
        ToolExecutionError: If the request fails for any reason.
    """
    url = f"{LINKEDIN_BASE_URL}{endpoint}"
    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    headers = {"Authorization": f"Bearer {token}"}

    async with httpx.AsyncClient() as client:
        try:
            response = await client.request(
                method, url, headers=headers, params=params, json=json_data
            )
            response.raise_for_status()
        except httpx.RequestError as e:
            raise ToolExecutionError(f"Failed to send request to LinkedIn API: {e}")

    return response


def _handle_linkedin_api_error(response: httpx.Response) -> None:
    """
    Handle errors from the LinkedIn API by mapping common status codes to ToolExecutionErrors.

    Args:
        response: The response object from the API request.

    Raises:
        ToolExecutionError: If the response contains an error status code.
    """
    status_code_map = {
        401: ToolExecutionError("Unauthorized: Invalid or expired token"),
        403: ToolExecutionError("Forbidden: User does not have Spotify Premium"),
        429: ToolExecutionError("Too Many Requests: Rate limit exceeded"),
    }

    if response.status_code in status_code_map:
        raise status_code_map[response.status_code]
    elif response.status_code >= 400:
        raise ToolExecutionError(f"Error: {response.status_code} - {response.text}")



"""
This is a simple example of how to use Arcade with CrewAI.
The example authenticates into the user's Gmail account, retrieves their 5 most recent emails, and summarizes them.

The example assumes the following:
1. You have an Arcade API key and have set the ARCADE_API_KEY environment variable.
2. You have an OpenAI API key and have set the OPENAI_API_KEY environment variable.
3. You have installed the necessary dependencies in the requirements.txt file: `pip install -r requirements.txt`

"""

from crewai import Agent, Crew, Task
from crewai.llm import LLM
from crewai_arcade import ArcadeToolManager

manager = ArcadeToolManager(default_user_id="user@example.com")
tools = manager.get_tools(tools=["Google.ListEmails"])

crew_agent = Agent(
    role="Main Agent",
    backstory="You are a helpful assistant",
    goal="Help the user with their requests",
    tools=tools,
    allow_delegation=False,
    verbose=True,
    llm=LLM(model="gpt-4o"),
)

task = Task(
    description="Get the 5 most recent emails from the user's inbox and summarize them and recommend a response for each.",
    expected_output="A bulleted list with a one sentence summary of each email and a recommended response to the email.",
    agent=crew_agent,
    tools=crew_agent.tools,
)

crew = Crew(
    agents=[crew_agent],
    tasks=[task],
    verbose=True,
    memory=True,
)

result = crew.kickoff()

print("\n\n\n ------------ Result ------------ \n\n\n")
print(result)



import re
from typing import Any, cast

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError
from serpapi import Client as SerpClient


def prepare_params(engine: str, **kwargs: Any) -> dict[str, Any]:
    """
    Prepares a parameters dictionary for the SerpAPI call.

    Parameters:
        engine: The engine name (e.g., "google", "google_finance").
        kwargs: Any additional parameters to include.

    Returns:
        A dictionary containing the base parameters plus any extras,
        excluding any parameters whose value is None.
    """
    params = {"engine": engine}
    params.update({k: v for k, v in kwargs.items() if v is not None})
    return params


def call_serpapi(context: ToolContext, params: dict) -> dict:
    """
    Execute a search query using the SerpAPI client and return the results as a dictionary.

    Args:
        context: The tool context containing required secrets.
        params: A dictionary of parameters for the SerpAPI search.

    Returns:
        The search results as a dictionary.
    """
    api_key = context.get_secret("SERP_API_KEY")
    client = SerpClient(api_key=api_key)
    try:
        search = client.search(params)
        return cast(dict[str, Any], search.as_dict())
    except Exception as e:
        # SerpAPI error messages sometimes contain the API key, so we need to sanitize it
        sanitized_e = re.sub(r"(api_key=)[^ &]+", r"\1***", str(e))
        raise ToolExecutionError(
            message="Failed to fetch search results",
            developer_message=sanitized_e,
        )


def extract_walmart_results(results: list[dict[str, Any]]) -> list[dict[str, Any]]:
    return [
        {
            "item_id": result.get("us_item_id"),
            "title": result.get("title"),
            "description": result.get("description"),
            "rating": result.get("rating"),
            "reviews_count": result.get("reviews"),
            "seller": {
                "id": result.get("seller_id"),
                "name": result.get("seller_name"),
            },
            "price": {
                "value": result.get("primary_offer", {}).get("offer_price"),
                "currency": result.get("primary_offer", {}).get("offer_currency"),
            },
            "link": result.get("product_page_url"),
        }
        for result in results
    ]


def get_walmart_last_page_integer(results: dict[str, Any]) -> int:
    try:
        return int(list(results["pagination"]["other_pages"].keys())[-1])
    except (KeyError, IndexError, ValueError):
        return 1


def extract_walmart_product_details(product: dict[str, Any]) -> dict[str, Any]:
    return {
        "item_id": product.get("us_item_id"),
        "product_type": product.get("product_type"),
        "title": product.get("title"),
        "description_html": product.get("short_description_html"),
        "rating": product.get("rating"),
        "reviews_count": product.get("reviews"),
        "seller": {
            "id": product.get("seller_id"),
            "name": product.get("seller_name"),
        },
        "manufacturer_name": product.get("manufacturer"),
        "price": {
            "value": product.get("price_map", {}).get("price"),
            "currency": product.get("price_map", {}).get("currency"),
            "previous_price": product.get("price_map", {}).get("was_price", {}).get("price"),
        },
        "link": product.get("product_page_url"),
        "variant_options": extract_walmart_variant_options(product.get("variant_swatches", [])),
    }


def extract_walmart_variant_options(variant_swatches: list[dict[str, Any]]) -> list[dict[str, Any]]:
    variants = []

    for variant_swatch in variant_swatches:
        variant_name = variant_swatch.get("name")
        if not variant_name:
            continue

        options = []

        for selection in variant_swatch.get("available_selections", []):
            selection_name = selection.get("name")
            if selection_name and selection_name not in options:
                options.append(selection_name)

        variants.append({variant_name: options})

    return variants



import asyncio
from collections import OrderedDict
from threading import Lock
from typing import Generic, TypeVar

from arcade_jira.constants import JIRA_CACHE_MAX_ITEMS

T = TypeVar("T")


class LRUCache(Generic[T]):
    def __init__(self, max_size: int):
        self.cache: OrderedDict[str, T] = OrderedDict()
        self.max_size = max_size
        self.thread_lock = Lock()
        self.async_lock = asyncio.Lock()

    # Thread-safe synchronous methods
    def get(self, key: str) -> T | None:
        with self.thread_lock:
            if key not in self.cache:
                return None

            value = self.cache.pop(key)
            self.cache[key] = value
            return value

    def set(self, key: str, value: T) -> None:
        with self.thread_lock:
            if key in self.cache:
                self.cache.pop(key)
            elif len(self.cache) >= self.max_size:
                self.cache.popitem(last=False)
            self.cache[key] = value

    # Async-safe methods
    async def async_get(self, key: str) -> T | None:
        async with self.async_lock:
            if key not in self.cache:
                return None

            value = self.cache.pop(key)
            self.cache[key] = value
            return value

    async def async_set(self, key: str, value: T) -> None:
        async with self.async_lock:
            if key in self.cache:
                self.cache.pop(key)
            elif len(self.cache) >= self.max_size:
                self.cache.popitem(last=False)
            self.cache[key] = value


CLOUD_ID_CACHE = LRUCache[str](max_size=JIRA_CACHE_MAX_ITEMS)
CLOUD_NAME_CACHE = LRUCache[str](max_size=JIRA_CACHE_MAX_ITEMS)
CLIENT_SEMAPHORE_CACHE = LRUCache[asyncio.Semaphore](max_size=JIRA_CACHE_MAX_ITEMS)


def get_cloud_id(auth_token: str) -> str | None:
    return CLOUD_ID_CACHE.get(auth_token)


def get_cloud_name(auth_token: str) -> str | None:
    return CLOUD_NAME_CACHE.get(auth_token)


def set_cloud_id(auth_token: str, cloud_id: str) -> None:
    CLOUD_ID_CACHE.set(auth_token, cloud_id)


def set_cloud_name(auth_token: str, cloud_name: str) -> None:
    CLOUD_NAME_CACHE.set(auth_token, cloud_name)


def get_jira_client_semaphore(auth_token: str) -> asyncio.Semaphore | None:
    return CLIENT_SEMAPHORE_CACHE.get(auth_token)


def set_jira_client_semaphore(auth_token: str, semaphore: asyncio.Semaphore) -> None:
    CLIENT_SEMAPHORE_CACHE.set(auth_token, semaphore)


async def async_get_cloud_id(auth_token: str) -> str | None:
    return await CLOUD_ID_CACHE.async_get(auth_token)


async def async_get_cloud_name(auth_token: str) -> str | None:
    return await CLOUD_NAME_CACHE.async_get(auth_token)


async def async_set_cloud_id(auth_token: str, cloud_id: str) -> None:
    await CLOUD_ID_CACHE.async_set(auth_token, cloud_id)


async def async_set_cloud_name(auth_token: str, cloud_name: str) -> None:
    await CLOUD_NAME_CACHE.async_set(auth_token, cloud_name)


async def async_get_jira_client_semaphore(auth_token: str) -> asyncio.Semaphore | None:
    return await CLIENT_SEMAPHORE_CACHE.async_get(auth_token)


async def async_set_jira_client_semaphore(auth_token: str, semaphore: asyncio.Semaphore) -> None:
    await CLIENT_SEMAPHORE_CACHE.async_set(auth_token, semaphore)



"""

This is an example of how to use Arcade with CrewAI.
The ArcadeToolManager allows you to handle both authorization and tool execution in a custom way.
This example demonstrates how to implement a custom auth handler and a custom tool execute handler.

The example assumes the following:
1. You have an Arcade API key and have set the ARCADE_API_KEY environment variable.
2. You have an OpenAI API key and have set the OPENAI_API_KEY environment variable.
3. You have installed the necessary dependencies in the requirements.txt file: `pip install -r requirements.txt`

"""

from typing import Any

from crewai import Agent, Crew, Task
from crewai.crews import CrewOutput
from crewai.llm import LLM
from crewai_arcade import ArcadeToolManager

USER_ID = "user@example.com"


def custom_auth_flow(
    manager: ArcadeToolManager, tool_name: str, **tool_input: dict[str, Any]
) -> Any:
    """Custom auth flow for the ArcadeToolManager

    This function is called when CrewAI needs to call a tool that requires authorization.
    Authorization is handled before executing the tool.
    This function overrides the ArcadeToolManager's default auth flow performed by ArcadeToolManager.authorize_tool
    """
    # Get authorization status
    auth_response = manager.authorize(tool_name, USER_ID)

    # If the user is not authorized for the tool,
    # then we need to handle the authorization before executing the tool
    if not manager.is_authorized(auth_response.id):
        print(f"Authorization required for tool: '{tool_name}' with inputs:")
        for input_name, input_value in tool_input.items():
            print(f"  {input_name}: {input_value}")
        # Handle authorization
        print(f"\nTo authorize, visit: {auth_response.url}")
        # Block until the user has completed the authorization
        auth_response = manager.wait_for_auth(auth_response)

        # Ensure authorization completed successfully
        if not manager.is_authorized(auth_response.id):
            raise ValueError(f"Authorization failed for {tool_name}. URL: {auth_response.url}")
    else:
        print(f"Authorization already granted for tool: '{tool_name}' with inputs:")
        for input_name, input_value in tool_input.items():
            print(f"  {input_name}: {input_value}")


def custom_execute_flow(
    manager: ArcadeToolManager, tool_name: str, **tool_input: dict[str, Any]
) -> Any:
    """Custom tool execution flow for the ArcadeToolManager

    This function is called when CrewAI needs to execute a tool after any authorization has been handled.
    This function overrides the ArcadeToolManager's default tool execution flow performed by ArcadeToolManager.execute_tool
    """
    print(f"Executing tool: '{tool_name}' with inputs:")
    for input_name, input_value in tool_input.items():
        print(f"  {input_name}: {input_value}")

    # Execute the tool
    response = manager._client.tools.execute(
        tool_name=tool_name,
        input=tool_input,
        user_id=USER_ID,
    )

    # Handle the tool error if it exists
    tool_error = response.output.error if response.output else None
    if tool_error:
        return str(tool_error)

    # Return the tool output if the tool was executed successfully
    if response.success:
        return response.output.value  # type: ignore[union-attr]

    # Return a failure message if the tool was not executed successfully
    return "Failed to call " + tool_name


def custom_tool_executor(
    manager: ArcadeToolManager, tool_name: str, **tool_input: dict[str, Any]
) -> Any:
    """Custom tool executor for the ArcadeToolManager

    ArcadeToolManager's default executor handles authorization and tool execution.
    This function overrides the default executor to handle authorization and tool execution in a custom way.
    """
    custom_auth_flow(manager, tool_name, **tool_input)
    return custom_execute_flow(manager, tool_name, **tool_input)


def main() -> CrewOutput:
    manager = ArcadeToolManager(
        executor=custom_tool_executor,
    )
    tools = manager.get_tools(tools=["Google.ListEmails"])

    crew_agent = Agent(
        role="Main Agent",
        backstory="You are a helpful assistant",
        goal="Help the user with their requests",
        tools=tools,
        allow_delegation=False,
        verbose=True,
        llm=LLM(model="gpt-4o"),
    )

    task = Task(
        description="Get the 5 most recent emails from the user's inbox and summarize them and recommend a response for each.",
        expected_output="A bulleted list with a one sentence summary of each email and a recommended response to the email.",
        agent=crew_agent,
        tools=crew_agent.tools,
    )

    crew = Crew(
        agents=[crew_agent],
        tasks=[task],
        verbose=True,
        memory=True,
    )

    result = crew.kickoff()
    return result


if __name__ == "__main__":
    result = main()
    print("\n\n\n ------------ Result ------------ \n\n\n")
    print(result)



import os

PROD_CLOUD_HOST = "cloud.arcade.dev"
PROD_ENGINE_HOST = "api.arcade.dev"
LOCALHOST = "localhost"
LOCAL_AUTH_CALLBACK_PORT = 9905

# The path to the directory containing the Arcade configuration files. Typically ~/.arcade
ARCADE_CONFIG_PATH = os.path.join(os.path.expanduser(os.getenv("ARCADE_WORK_DIR", "~")), ".arcade")

# The path to the file containing the user's Arcade-related credentials (e.g., ARCADE_API_KEY).
CREDENTIALS_FILE_PATH = os.path.join(ARCADE_CONFIG_PATH, "credentials.yaml")

_style_block = b"""
<link rel="icon" href="https://cdn.arcade.dev/favicons/favicon.ico" sizes="any">
<link rel="apple-touch-icon" href="https://cdn.arcade.dev/favicons/apple-touch-icon.png">
<link rel="icon" type="image/png" sizes="32x32" href="https://cdn.arcade.dev/favicons/favicon-32x32.png">
<link rel="icon" type="image/png" sizes="16x16" href="https://cdn.arcade.dev/favicons/favicon-16x16.png">
<link rel="apple-touch-icon" sizes="180x180" href="https://cdn.arcade.dev/favicons/apple-touch-icon.png">

<style>
    body {
        display: flex;
        justify-content: center;
        align-items: center;
        height: 100vh;
        margin: 0;
        background: linear-gradient(135deg, #1a1a1a, #0f0f0f);
        font-family: Arial, sans-serif;
    }

    .container {
        background-color: #333;
        padding: 40px;
        border-radius: 8px;
        box-shadow: 0 4px 8px rgba(0, 0, 0, 0.2);
        width: 300px;
    }

    .container h2 {
        color: #fff;
        margin-bottom: 20px;
        text-align: center;
    }

    .container label {
        display: block;
        color: #bbb;
        margin-bottom: 5px;
        font-size: 14px;
    }

    .container input[type="text"],
    .container input[type="password"] {
        width: 100%;
        padding: 10px;
        margin-bottom: 15px;
        border: none;
        border-radius: 4px;
        background-color: #444;
        color: #ddd;
        font-size: 16px;
        box-sizing: border-box;
    }

    .container input[type="text"]::placeholder,
    .container input[type="password"]::placeholder {
        color: #aaa;
    }

    .container input[type="submit"] {
        width: 100%;
        padding: 10px;
        border: none;
        border-radius: 4px;
        background-color: #ED155D;
        color: #fff;
        font-size: 16px;
        cursor: pointer;
        transition: background-color 0.3s ease;
    }

    .container input[type="submit"]:hover {
        background-color: #C0104A;
    }

    .message {
        background-color: #1e1e1e;
        padding: 10px;
        border-radius: 4px;
        margin-bottom: 15px;
        font-size: 14px;
        text-align: center;
    }
    .info {
        color: #fff;
    }

    .error {
        color: #ff4d4d;
    }

    .logo {
        display: block;
        max-width: 100%;
        max-height: 90px;
        margin: 0 auto 20px;
    }
</style>
"""

LOGIN_SUCCESS_HTML = (
    b"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Success!</title>
    """
    + _style_block
    + b"""
</head>
<body>
    <div class="container">
        <img src="https://cdn.arcade.dev/logos/a-icon.png" alt="Arcade logo" class="logo">
        <h2>Log in to Arcade CLI</h2>
        <p class="message info">Success! You can close this window.</p>
    </div>
</body>
</html>
"""
)

LOGIN_FAILED_HTML = (
    b"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Login failed</title>
    """
    + _style_block
    + b"""
</head>
<body>
    <div class="container">
        <img src="https://cdn.arcade.dev/logos/a-icon.png" alt="Arcade logo" class="logo">
        <h2>Log in to Arcade CLI</h2>
        <p class="message error">Something went wrong. Please close this window and try again.</p>
    </div>
</body>
</html>
"""
)



class SlackToolkitError(Exception):
    """Base class for all Slack toolkit errors."""


class PaginationTimeoutError(SlackToolkitError):
    """Raised when a timeout occurs during pagination."""

    def __init__(self, timeout_seconds: int):
        self.timeout_seconds = timeout_seconds
        super().__init__(f"The pagination process timed out after {timeout_seconds} seconds.")



from enum import Enum


class Corpora(str, Enum):
    """
    Bodies of items (files/documents) to which the query applies.
    Prefer 'user' or 'drive' to 'allDrives' for efficiency.
    By default, corpora is set to 'user'.
    """

    USER = "user"
    DOMAIN = "domain"
    DRIVE = "drive"
    ALL_DRIVES = "allDrives"


class OrderBy(str, Enum):
    """
    Sort keys for ordering files in Google Drive.
    Each key has both ascending and descending options.
    """

    CREATED_TIME = (
        # When the file was created (ascending)
        "createdTime"
    )
    CREATED_TIME_DESC = (
        # When the file was created (descending)
        "createdTime desc"
    )
    FOLDER = (
        # The folder ID, sorted using alphabetical ordering (ascending)
        "folder"
    )
    FOLDER_DESC = (
        # The folder ID, sorted using alphabetical ordering (descending)
        "folder desc"
    )
    MODIFIED_BY_ME_TIME = (
        # The last time the file was modified by the user (ascending)
        "modifiedByMeTime"
    )
    MODIFIED_BY_ME_TIME_DESC = (
        # The last time the file was modified by the user (descending)
        "modifiedByMeTime desc"
    )
    MODIFIED_TIME = (
        # The last time the file was modified by anyone (ascending)
        "modifiedTime"
    )
    MODIFIED_TIME_DESC = (
        # The last time the file was modified by anyone (descending)
        "modifiedTime desc"
    )
    NAME = (
        # The name of the file, sorted using alphabetical ordering (e.g., 1, 12, 2, 22) (ascending)
        "name"
    )
    NAME_DESC = (
        # The name of the file, sorted using alphabetical ordering (e.g., 1, 12, 2, 22) (descending)
        "name desc"
    )
    NAME_NATURAL = (
        # The name of the file, sorted using natural sort ordering (e.g., 1, 2, 12, 22) (ascending)
        "name_natural"
    )
    NAME_NATURAL_DESC = (
        # The name of the file, sorted using natural sort ordering (e.g., 1, 2, 12, 22) (descending)
        "name_natural desc"
    )
    QUOTA_BYTES_USED = (
        # The number of storage quota bytes used by the file (ascending)
        "quotaBytesUsed"
    )
    QUOTA_BYTES_USED_DESC = (
        # The number of storage quota bytes used by the file (descending)
        "quotaBytesUsed desc"
    )
    RECENCY = (
        # The most recent timestamp from the file's date-time fields (ascending)
        "recency"
    )
    RECENCY_DESC = (
        # The most recent timestamp from the file's date-time fields (descending)
        "recency desc"
    )
    SHARED_WITH_ME_TIME = (
        # When the file was shared with the user, if applicable (ascending)
        "sharedWithMeTime"
    )
    SHARED_WITH_ME_TIME_DESC = (
        # When the file was shared with the user, if applicable (descending)
        "sharedWithMeTime desc"
    )
    STARRED = (
        # Whether the user has starred the file (ascending)
        "starred"
    )
    STARRED_DESC = (
        # Whether the user has starred the file (descending)
        "starred desc"
    )
    VIEWED_BY_ME_TIME = (
        # The last time the file was viewed by the user (ascending)
        "viewedByMeTime"
    )
    VIEWED_BY_ME_TIME_DESC = (
        # The last time the file was viewed by the user (descending)
        "viewedByMeTime desc"
    )


class DocumentFormat(str, Enum):
    MARKDOWN = "markdown"
    HTML = "html"
    GOOGLE_API_JSON = "google_api_json"



import os
from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel, ConfigDict, ValidationError


class BaseConfig(BaseModel):
    model_config = ConfigDict(extra="ignore")


class ApiConfig(BaseConfig):
    """
    Arcade API configuration.
    """

    key: str
    """
    Arcade API key.
    """
    version: str = "v1"
    """
    Arcade API version.
    """


class UserConfig(BaseConfig):
    """
    Arcade user configuration.
    """

    email: str | None = None
    """
    User email.
    """


class Config(BaseConfig):
    """
    Configuration for Arcade.
    """

    api: ApiConfig
    """
    Arcade API configuration.
    """
    user: UserConfig | None = None
    """
    Arcade user configuration.
    """

    def __init__(self, **data: Any):
        super().__init__(**data)

    @classmethod
    def get_config_dir_path(cls) -> Path:
        """
        Get the path to the Arcade configuration directory.
        """
        config_path = os.getenv("ARCADE_WORK_DIR") or Path.home() / ".arcade"
        return Path(config_path).resolve()

    @classmethod
    def get_config_file_path(cls) -> Path:
        """
        Get the path to the Arcade configuration file.
        """
        return cls.get_config_dir_path() / "credentials.yaml"

    @classmethod
    def ensure_config_dir_exists(cls) -> None:
        """
        Create the configuration directory if it does not exist.
        """
        config_dir = Config.get_config_dir_path()
        if not config_dir.exists():
            config_dir.mkdir(parents=True, exist_ok=True)

    @classmethod
    def load_from_file(cls) -> "Config":
        """
        Load the configuration from the YAML file in the configuration directory.

        If no configuration file exists, this method will create a new one with default values.
        The default configuration includes:
        - An empty API configuration
        - A default Engine configuration (host: "api.arcade.dev", port: None, tls: True)
        - No user configuration

        Returns:
            Config: The loaded or newly created configuration.

        Raises:
            ValueError: If the existing configuration file is invalid.
        """
        cls.ensure_config_dir_exists()

        config_file_path = cls.get_config_file_path()

        if not config_file_path.exists():
            # Create a file using the default configuration
            default_config = cls.model_construct(api=ApiConfig.model_construct())
            default_config.save_to_file()

        config_data = yaml.safe_load(config_file_path.read_text())

        if config_data is None:
            raise ValueError(
                "Invalid credentials.yaml file. Please ensure it is a valid YAML file."
            )

        if "cloud" not in config_data:
            raise ValueError("Invalid credentials.yaml file. Expected a 'cloud' key.")

        try:
            return cls(**config_data["cloud"])
        except ValidationError as e:
            # Get only the errors with {type:missing} and combine them
            # into a nicely-formatted string message.
            # Any other errors without {type:missing} should just be str()ed
            missing_field_errors = [
                ".".join(map(str, error["loc"]))
                for error in e.errors()
                if error["type"] == "missing"
            ]
            other_errors = [str(error) for error in e.errors() if error["type"] != "missing"]

            missing_field_errors_str = ", ".join(missing_field_errors)
            other_errors_str = "\n".join(other_errors)

            pretty_str: str = "Invalid Arcade configuration."
            if missing_field_errors_str:
                pretty_str += f"\nMissing fields: {missing_field_errors_str}\n"
            if other_errors_str:
                pretty_str += f"\nOther errors:\n{other_errors_str}"

            raise ValueError(pretty_str) from e

    def save_to_file(self) -> None:
        """
        Save the configuration to the YAML file in the configuration directory.
        """
        Config.ensure_config_dir_exists()
        config_file_path = Config.get_config_file_path()
        config_file_path.write_text(yaml.dump(self.model_dump()))



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from firecrawl import FirecrawlApp


# TODO: Support scrapeOptions.
@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def crawl_website(
    context: ToolContext,
    url: Annotated[str, "URL to crawl"],
    exclude_paths: Annotated[list[str] | None, "URL patterns to exclude from the crawl"] = None,
    include_paths: Annotated[list[str] | None, "URL patterns to include in the crawl"] = None,
    max_depth: Annotated[int, "Maximum depth to crawl relative to the entered URL"] = 2,
    ignore_sitemap: Annotated[bool, "Ignore the website sitemap when crawling"] = True,
    limit: Annotated[int, "Limit the number of pages to crawl"] = 10,
    allow_backward_links: Annotated[
        bool,
        "Enable navigation to previously linked pages and enable crawling "
        "sublinks that are not children of the 'url' input parameter.",
    ] = False,
    allow_external_links: Annotated[bool, "Allow following links to external websites"] = False,
    webhook: Annotated[
        str | None,
        "The URL to send a POST request to when the crawl is started, updated and completed.",
    ] = None,
    async_crawl: Annotated[bool, "Run the crawl asynchronously"] = True,
) -> Annotated[dict[str, Any], "Crawl status and data"]:
    """
    Crawl a website using Firecrawl. If the crawl is asynchronous, then returns the crawl ID.
    If the crawl is synchronous, then returns the crawl data.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    params = {
        "limit": limit,
        "excludePaths": exclude_paths or [],
        "includePaths": include_paths or [],
        "maxDepth": max_depth,
        "ignoreSitemap": ignore_sitemap,
        "allowBackwardLinks": allow_backward_links,
        "allowExternalLinks": allow_external_links,
    }
    if webhook:
        params["webhook"] = webhook

    if async_crawl:
        response = app.async_crawl_url(url, params=params)
        response.pop("url", None)  # Remove 'url' as it's an API endpoint

        if response["success"]:
            response["status"] = await get_crawl_status(context, response["id"])
            response["llm_instructions"] = (
                "You have the ability to get crawl status, cancel a crawl, "
                "and get a crawl's data. Inform the user that you have these capabilities. "
                "Inform the user that they should let you know if they want you to perform any "
                "of these actions."
            )

    else:
        response = app.crawl_url(url, params=params)

    return dict(response)


@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def get_crawl_status(
    context: ToolContext,
    crawl_id: Annotated[str, "The ID of the crawl job"],
) -> Annotated[dict[str, Any], "Crawl status information"]:
    """
    Get the status of a Firecrawl 'crawl' that is either in progress or recently completed.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    crawl_status = app.check_crawl_status(crawl_id)

    crawl_status.pop("data", None)  # Remove 'data' if it exists
    crawl_status.pop("next", None)  # Remove 'next' as it's an API endpoint

    return dict(crawl_status)


# TODO: Support responses greater than 10 MB. If the response is greater than 10 MB,
#       then the Firecrawl API response will have a next_url field.
@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def get_crawl_data(
    context: ToolContext,
    crawl_id: Annotated[str, "The ID of the crawl job"],
) -> Annotated[dict[str, Any], "Crawl data information"]:
    """
    Get the data of a Firecrawl 'crawl' that is either in progress or recently completed.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    crawl_data = app.check_crawl_status(crawl_id)

    return dict(crawl_data)


@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def cancel_crawl(
    context: ToolContext,
    crawl_id: Annotated[str, "The ID of the asynchronous crawl job to cancel"],
) -> Annotated[dict[str, Any], "Cancellation status information"]:
    """
    Cancel an asynchronous crawl job that is in progress using the Firecrawl API.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    cancellation_status = app.cancel_crawl(crawl_id)

    return dict(cancellation_status)



from datetime import timedelta

from arcade_evals import (
    BinaryCritic,
    DatetimeCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

from arcade_outlook_mail import (
    list_emails,
    list_emails_by_property,
    list_emails_in_folder,
)
from arcade_outlook_mail.enums import WellKnownFolderNames
from evals.additional_messages import (
    list_emails_with_pagination_token_additional_messages,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_tool(list_emails, "OutlookMail")
catalog.add_tool(list_emails_in_folder, "OutlookMail")
catalog.add_tool(list_emails_by_property, "OutlookMail")


@tool_eval()
def outlook_mail_read_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Outlook Mail tools."""
    suite = EvalSuite(
        name="Outlook Mail Tools Evaluation",
        system_message=("You are an AI that has access to tools to send, read, and write emails."),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List emails in mailbox",
        user_message="get my five most recent emails",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails,
                args={"limit": 5},
            )
        ],
        critics=[
            BinaryCritic(critic_field="limit", weight=1.0),
        ],
    )

    suite.add_case(
        name="List emails in mailbox with pagination token",
        user_message="get the next 3",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails,
                args={
                    "limit": 3,
                    "pagination_token": "https://graph.microsoft.com/v1.0/me/messages?%24count=true&%24orderby=receivedDateTime+DESC&%24select=bccRecipients%2cbody%2cccRecipients%2cconversationId%2cconversationIndex%2cflag%2cfrom%2chasAttachments%2cimportance%2cisDraft%2cisRead%2creceivedDateTime%2creplyTo%2csubject%2ctoRecipients%2cwebLink&%24top=1&%24skip=1",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="limit", weight=0.2),
            BinaryCritic(critic_field="pagination_token", weight=0.8),
        ],
        additional_messages=list_emails_with_pagination_token_additional_messages,
    )

    suite.add_case(
        name="List emails in well-known folder",
        user_message="summarize my inbox",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_in_folder,
                args={
                    "well_known_folder_name": WellKnownFolderNames.INBOX,
                    "folder_id": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="well_known_folder_name", weight=0.5),
            BinaryCritic(critic_field="folder_id", weight=0.5),
        ],
    )

    suite.add_case(
        name="List emails in folder by id",
        user_message="get 5 from folder AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoALgAAAyXxSd3UxTpCkDpGouEg0JMBAFuxokOLZRtDncM4",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_in_folder,
                args={
                    "well_known_folder_name": None,
                    "folder_id": "AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoALgAAAyXxSd3UxTpCkDpGouEg0JMBAFuxokOLZRtDncM4",
                    "limit": 5,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="well_known_folder_name", weight=0.4),
            BinaryCritic(critic_field="folder_id", weight=0.4),
            BinaryCritic(critic_field="limit", weight=0.2),
        ],
    )

    return suite


@tool_eval()
def outlook_mail_list_emails_by_property_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Outlook Mail tools."""
    suite = EvalSuite(
        name="Outlook Mail Tools Evaluation",
        system_message=("You are an AI that has access to tools to send, read, and write emails."),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List emails by subject",
        user_message="get all emails that talk about The Green Bottle",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_by_property,
                args={
                    "property": "subject",
                    "operator": "contains",
                    "value": "The Green Bottle",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="property", weight=1 / 3),
            BinaryCritic(critic_field="operator", weight=1 / 3),
            BinaryCritic(critic_field="value", weight=1 / 3),
        ],
    )

    suite.extend_case(
        name="List emails by thread",
        user_message="get all emails in my thread 1k2jh324h92f24krjb34mtb43kj4bk3tmn34b3k4nnm3tb34mntb34mntb3m4bt3mn4bt3mn4btmnb34tmnb3t4mnb==34tkjh",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_by_property,
                args={
                    "property": "conversationId",
                    "operator": "eq",
                    "value": "1k2jh324h92f24krjb34mtb43kj4bk3tmn34b3k4nnm3tb34mntb34mntb3m4bt3mn4bt3mn4btmnb34tmnb3t4mnb==34tkjh",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="property", weight=1 / 3),
            BinaryCritic(critic_field="operator", weight=1 / 3),
            BinaryCritic(critic_field="value", weight=1 / 3),
        ],
    )

    suite.extend_case(
        name="List emails by date",
        user_message="Today is May 1st 2025. Get all emails that are a year old or older",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_by_property,
                args={
                    "property": "receivedDateTime",
                    "operator": "le",
                    "value": "2024-05-01T00:00:00Z",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="property", weight=1 / 3),
            BinaryCritic(critic_field="operator", weight=1 / 3),
            DatetimeCritic(critic_field="value", weight=1 / 3, tolerance=timedelta(days=1)),
        ],
    )

    suite.extend_case(
        name="List emails by sender",
        user_message="get all of my correspondence with the folks over at arcade.dev",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_by_property,
                args={
                    "property": "sender/emailAddress/address",
                    "operator": "contains",
                    "value": "arcade.dev",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="property", weight=1 / 3),
            BinaryCritic(critic_field="operator", weight=1 / 3),
            BinaryCritic(critic_field="value", weight=1 / 3),
        ],
    )

    return suite



from arcade_tdk.errors import ToolExecutionError


class SalesforceToolExecutionError(ToolExecutionError):
    def __init__(self, errors: list[str], message: str = "Failed to execute Salesforce tool"):
        self.message = message
        self.errors = errors
        exc_message = f"{message}. Errors: {errors}"
        super().__init__(
            message=exc_message,
            developer_message=exc_message,
        )


class ResourceNotFoundError(SalesforceToolExecutionError):
    def __init__(self, errors: list[str]):
        super().__init__(message="Resource not found", errors=errors)


class BadRequestError(SalesforceToolExecutionError):
    def __init__(self, errors: list[str]):
        super().__init__(message="Bad request", errors=errors)



from typing import Annotated, cast

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Atlassian

from arcade_jira.client import JiraClient


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def get_transition_by_id(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue"],
    transition_id: Annotated[str, "The ID of the transition"],
) -> Annotated[dict, "The transition data"]:
    """Get a transition by its ID."""
    if not transition_id:
        return {"error": "The transition ID is required."}
    if not transition_id.isdigit():
        return {"error": "The transition ID must be a numeric string."}

    client = JiraClient(context.get_auth_token_or_empty())
    response = await client.get(
        f"/issue/{issue}/transitions",
        params={
            "transitionId": transition_id,
        },
    )
    transitions = response["transitions"]

    if len(transitions) == 0:
        return {
            "error": (
                f"No transition found for the issue '{issue}' with ID '{transition_id}'. "
                "To get all transitions available for the issue, use the "
                f"`Jira.{get_transitions_available_for_issue.__tool_name__}` tool."
            ),
        }

    if len(transitions) == 1:
        return {"transition": transitions[0]}

    return {
        "error": f"Multiple transitions found for the issue '{issue}' with ID '{transition_id}'.",
        "transitions": transitions,
    }


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def get_transitions_available_for_issue(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue"],
) -> Annotated[dict, "The transitions available and the issue's current status"]:
    """Get the transitions available for an existing Jira issue."""
    from arcade_jira.tools.issues import get_issue_by_id  # Avoid circular import

    client = JiraClient(context.get_auth_token_or_empty())
    issue_data = await get_issue_by_id(context, issue)
    if issue_data.get("error"):
        return cast(dict, issue_data)
    response = await client.get(
        f"/issue/{issue_data['issue']['id']}/transitions",
        params={
            "expand": "transitions.fields",
        },
    )
    return {
        "issue": {
            "id": issue_data["issue"]["id"],
            "key": issue_data["issue"]["key"],
            "current_status": issue_data["issue"]["status"],
        },
        "transitions_available": response["transitions"],
    }


@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to get the transitions available for the issue
            "write:jira-work",  # Needed to transition the issue
        ],
    ),
)
async def get_transition_by_status_name(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue"],
    transition: Annotated[str, "The name of the transition status"],
) -> Annotated[dict, "The transition data, including screen fields available"]:
    """Get a transition available for an issue by the transition name.

    The response will contain screen fields available for the transition, if any.
    """
    transitions = await get_transitions_available_for_issue(context, issue)
    for available_transition in transitions["transitions_available"]:
        if available_transition["name"].casefold() == transition.casefold():
            return {"issue": issue, "transition": available_transition}
    return {
        "error": f"Transition '{transition}' not found for the issue '{issue}'",
        "transitions_available": transitions["transitions_available"],
    }


@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to get the transition ID by name
            "write:jira-work",  # Needed to transition the issue
        ],
    ),
)
async def transition_issue_to_new_status(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue"],
    transition: Annotated[
        str,
        "The transition to perform. Provide the transition ID or its name (case insensitive).",
    ],
) -> Annotated[dict, "The updated issue"]:
    """Transition a Jira issue to a new status."""
    client = JiraClient(context.get_auth_token_or_empty())

    # Try to get the transition by ID first
    response = await get_transition_by_id(context, issue, transition)

    # If the transition is not found by ID, try to get it by name
    if response.get("error"):
        response = await get_transition_by_status_name(context, issue, transition)
        if response.get("error"):
            return cast(dict, response)

    transition_id = response["transition"]["id"]
    transition_name = response["transition"]["name"]

    # The /issue/issue_id/transitions endpoint returns a 204 No Content in case of success
    await client.post(
        f"/issue/{issue}/transitions",
        json_data={
            "transition": {"id": transition_id},
        },
    )

    return {
        "status": "success",
        "message": f"Issue '{issue}' successfully transitioned to '{transition_name}'.",
    }



import re
from typing import Any, cast

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError
from serpapi import Client as SerpClient

from arcade_google_shopping.constants import (
    DEFAULT_GOOGLE_COUNTRY,
    DEFAULT_GOOGLE_LANGUAGE,
)
from arcade_google_shopping.exceptions import CountryNotFoundError, LanguageNotFoundError
from arcade_google_shopping.google_data import COUNTRY_CODES, LANGUAGE_CODES


def prepare_params(engine: str, **kwargs: Any) -> dict[str, Any]:
    """
    Prepares a parameters dictionary for the SerpAPI call.

    Parameters:
        engine: The engine name (e.g., "google", "google_finance").
        kwargs: Any additional parameters to include.

    Returns:
        A dictionary containing the base parameters plus any extras,
        excluding any parameters whose value is None.
    """
    params = {"engine": engine}
    params.update({k: v for k, v in kwargs.items() if v is not None})
    return params


def call_serpapi(context: ToolContext, params: dict) -> dict:
    """
    Execute a search query using the SerpAPI client and return the results as a dictionary.

    Args:
        context: The tool context containing required secrets.
        params: A dictionary of parameters for the SerpAPI search.

    Returns:
        The search results as a dictionary.
    """
    api_key = context.get_secret("SERP_API_KEY")
    client = SerpClient(api_key=api_key)
    try:
        search = client.search(params)
        return cast(dict[str, Any], search.as_dict())
    except Exception as e:
        # SerpAPI error messages sometimes contain the API key, so we need to sanitize it
        sanitized_e = re.sub(r"(api_key=)[^ &]+", r"\1***", str(e))
        raise ToolExecutionError(
            message="Failed to fetch search results",
            developer_message=sanitized_e,
        )


def default_language_code(default_service_language_code: str | None = None) -> str | None:
    if isinstance(default_service_language_code, str):
        return default_service_language_code.lower()
    elif isinstance(DEFAULT_GOOGLE_LANGUAGE, str):
        return DEFAULT_GOOGLE_LANGUAGE.lower()
    return None


def default_country_code(default_service_country_code: str | None = None) -> str | None:
    if isinstance(default_service_country_code, str):
        return default_service_country_code.lower()
    elif isinstance(DEFAULT_GOOGLE_COUNTRY, str):
        return DEFAULT_GOOGLE_COUNTRY.lower()
    return None


def resolve_language_code(
    language_code: str | None = None,
    default_service_language_code: str | None = None,
) -> str | None:
    language_code = language_code or default_language_code(default_service_language_code)

    if isinstance(language_code, str):
        language_code = language_code.lower()
        if language_code not in LANGUAGE_CODES:
            raise LanguageNotFoundError(language_code)

    return language_code


def resolve_country_code(
    country_code: str | None = None,
    default_service_country_code: str | None = None,
) -> str | None:
    country_code = country_code or default_country_code(default_service_country_code)

    if isinstance(country_code, str):
        country_code = country_code.lower()
        if country_code not in COUNTRY_CODES:
            raise CountryNotFoundError(country_code)

    return country_code


def extract_shopping_results(results: list[dict[str, Any]]) -> list[dict[str, Any]]:
    return [
        {
            "title": result.get("title"),
            "direct_link": result.get("link"),
            "google_link": result.get("product_link"),
            "source": result.get("source"),
            "price": result.get("price"),
            "product_rating": result.get("rating"),
            "product_reviews": result.get("reviews"),
            "store_rating": result.get("store_rating"),
            "store_reviews": result.get("store_reviews"),
            "delivery": result.get("delivery"),
        }
        for result in results
    ]



import pytest

from arcade_notion_toolkit.utils import simplify_search_result


@pytest.mark.parametrize(
    "item, expected_title",
    [
        # Case 1: Database object with top-level "title"
        (
            {
                "id": "db1",
                "object": "database",
                "title": [{"plain_text": "Database Title"}],
                "parent": {"type": "workspace", "workspace": True},
                "created_time": "2021-01-01T00:00:00.000Z",
                "last_edited_time": "2021-01-02T00:00:00.000Z",
                "url": "https://notion.so/database/db1",
                "public_url": "https://notion.so/database/public/db1",
            },
            "Database Title",
        ),
        # Case 2: Page object with properties "Title"
        (
            {
                "id": "page1",
                "object": "page",
                "properties": {
                    "Income Item": {
                        "id": "title",
                        "title": [
                            {
                                "annotations": {
                                    "bold": False,
                                    "code": False,
                                    "color": "default",
                                    "italic": False,
                                    "strikethrough": False,
                                    "underline": False,
                                },
                                "href": None,
                                "plain_text": "Page title with database parent",
                                "text": {
                                    "content": "Page title with database parent",
                                    "link": None,
                                },
                                "type": "text",
                            }
                        ],
                        "type": "title",
                    },
                },
                "parent": {"database_id": "db1"},
                "created_time": "2021-01-03T00:00:00.000Z",
                "last_edited_time": "2021-01-04T00:00:00.000Z",
                "url": "https://notion.so/page/page1",
                "public_url": "https://notion.so/page/public/page1",
            },
            "Page title with database parent",
        ),
        # Case 3: Page object with properties "title"
        (
            {
                "id": "page2",
                "object": "page",
                "properties": {"title": {"title": [{"plain_text": "Page Title from title prop"}]}},
                "parent": {"page_id": "parent_id"},
                "created_time": "2021-01-05T00:00:00.000Z",
                "last_edited_time": "2021-01-06T00:00:00.000Z",
                "url": "https://notion.so/page/page2",
                "public_url": "https://notion.so/page/public/page2",
            },
            "Page Title from title prop",
        ),
    ],
)
def test_simplify_search_result(item, expected_title):
    simplified = simplify_search_result(item)
    assert simplified["title"] == expected_title
    assert simplified["id"] == item.get("id")
    assert simplified["object"] == item.get("object")
    assert simplified["parent"] == item.get("parent")
    assert simplified["created_time"] == item.get("created_time")
    assert simplified["last_edited_time"] == item.get("last_edited_time")
    assert simplified["url"] == item.get("url")
    assert simplified["public_url"] == item.get("public_url")



import pytest
from msgraph.generated.models.email_address import EmailAddress as GraphEmailAddress
from msgraph.generated.models.message import Message as GraphMessage
from msgraph.generated.models.recipient import Recipient as GraphRecipient

from arcade_microsoft.outlook_mail.message import Message, Recipient


# Dummy classes to simulate SDK objects
class DummyBody:
    def __init__(self, content):
        self.content = content


class DummyFlagStatus:
    def __init__(self, value):
        self.value = value


class DummyImportance:
    def __init__(self, value):
        self.value = value


class DummyDueDateTime:
    def __init__(self, date_time):
        self.date_time = date_time


class DummyFlag:
    def __init__(self, flag_status, due_date_time):
        self.flag_status = DummyFlagStatus(flag_status)
        self.due_date_time = DummyDueDateTime(due_date_time)


class DummyDateTime:
    def __init__(self, date_str):
        self.date_str = date_str

    def isoformat(self):
        return self.date_str


def make_graph_recipient(rec_data):
    recipient = GraphRecipient()
    recipient.email_address = GraphEmailAddress()
    recipient.email_address.address = rec_data["email_address"]
    recipient.email_address.name = rec_data.get("name", "")
    return recipient


@pytest.mark.parametrize(
    "input_data, expected",
    [
        (
            {
                "body_content": "<p>Hello <b>world</b></p>",
                "subject": "Test subject",
                "conversation_id": "conv-1",
                "conversation_index": "conv-index",
                "flag_status": "flagged",
                "due_date_time": "2021-01-01T10:00:00",
                "has_attachments": False,
                "importance": "high",
                "is_read": True,
                "received_date_time": "2021-01-02T00:00:00",
                "web_link": "http://example.com",
                "is_draft": False,
                "message_id": "1234",
                "to_recipients": [{"email_address": "to@example.com", "name": "ToName"}],
                "cc_recipients": [{"email_address": "cc@example.com", "name": "CcName"}],
                "bcc_recipients": [{"email_address": "bcc@example.com", "name": "BccName"}],
                "reply_to": [{"email_address": "reply@example.com", "name": "ReplyName"}],
                "from_": {"email_address": "from@example.com", "name": "FromName"},
                "conversation_index_bytes": False,
            },
            {
                "body": "Hello world",
                "subject": "Test subject",
                "conversation_id": "conv-1",
                "conversation_index": "conv-index",
                "flag": {"flag_status": "flagged", "due_date_time": "2021-01-01T10:00:00"},
                "has_attachments": False,
                "importance": "high",
                "is_read": True,
                "received_date_time": "2021-01-02T00:00:00",
                "web_link": "http://example.com",
                "is_draft": False,
                "message_id": "1234",
                "to_recipients": [{"email_address": "to@example.com", "name": "ToName"}],
                "cc_recipients": [{"email_address": "cc@example.com", "name": "CcName"}],
                "bcc_recipients": [{"email_address": "bcc@example.com", "name": "BccName"}],
                "reply_to": [{"email_address": "reply@example.com", "name": "ReplyName"}],
                "from_": {"email_address": "from@example.com", "name": "FromName"},
            },
        ),
        (
            {
                "body_content": "<p>Sample <i>email</i> message</p>",
                "subject": "Another subject",
                "conversation_id": "conv-2",
                "conversation_index": b"byte-index",
                "flag_status": "notFlaged",
                "due_date_time": "",
                "has_attachments": False,
                "importance": "low",
                "is_read": False,
                "received_date_time": "",
                "web_link": "",
                "is_draft": True,
                "message_id": "5678",
                "to_recipients": [{"email_address": "user1@example.com", "name": "User1"}],
                "cc_recipients": [],
                "bcc_recipients": [],
                "reply_to": [],
                "from_": {"email_address": "sender@example.com", "name": "Sender"},
                "conversation_index_bytes": True,
            },
            {
                "body": "Sample email message",
                "subject": "Another subject",
                "conversation_id": "conv-2",
                "conversation_index": "byte-index",
                "flag": {"flag_status": "notFlaged", "due_date_time": ""},
                "has_attachments": False,
                "importance": "low",
                "is_read": False,
                "received_date_time": "",
                "web_link": "",
                "is_draft": True,
                "message_id": "5678",
                "to_recipients": [{"email_address": "user1@example.com", "name": "User1"}],
                "cc_recipients": [],
                "bcc_recipients": [],
                "reply_to": [],
                "from_": {"email_address": "sender@example.com", "name": "Sender"},
            },
        ),
    ],
)
def test_message_conversion(input_data, expected):
    # Set up sdk message
    sdk_message = GraphMessage()
    sdk_message.body = (
        DummyBody(input_data["body_content"]) if "body_content" in input_data else None
    )
    sdk_message.subject = input_data["subject"]
    sdk_message.conversation_id = input_data["conversation_id"]
    sdk_message.conversation_index = input_data["conversation_index"]
    sdk_message.flag = (
        DummyFlag(input_data["flag_status"], input_data["due_date_time"])
        if "flag_status" in input_data
        else None
    )
    sdk_message.has_attachments = input_data["has_attachments"]
    sdk_message.importance = DummyImportance(input_data["importance"])
    sdk_message.is_read = input_data["is_read"]
    sdk_message.received_date_time = (
        DummyDateTime(input_data["received_date_time"])
        if input_data["received_date_time"]
        else None
    )
    sdk_message.web_link = input_data["web_link"]
    sdk_message.is_draft = input_data["is_draft"]
    sdk_message.id = input_data["message_id"]
    sdk_message.to_recipients = [make_graph_recipient(r) for r in input_data["to_recipients"]]
    sdk_message.cc_recipients = [make_graph_recipient(r) for r in input_data["cc_recipients"]]
    sdk_message.bcc_recipients = [make_graph_recipient(r) for r in input_data["bcc_recipients"]]
    sdk_message.reply_to = [make_graph_recipient(r) for r in input_data["reply_to"]]
    sdk_message.from_ = make_graph_recipient(input_data["from_"])

    # Convert to Arcade Message type
    message = Message.from_sdk(sdk_message)

    # Ensure conversion is correct
    assert message.body == expected["body"], "Body conversion mismatch"
    assert message.subject == expected["subject"]
    assert message.conversation_id == expected["conversation_id"]
    assert message.conversation_index == expected["conversation_index"]
    assert message.flag == expected["flag"]
    assert message.has_attachments == expected["has_attachments"]
    assert message.importance == expected["importance"]
    assert message.is_read == expected["is_read"]
    assert message.received_date_time == expected["received_date_time"]
    assert message.web_link == expected["web_link"]
    assert message.is_draft == expected["is_draft"]
    assert message.message_id == expected["message_id"]
    assert message.from_.email_address == expected["from_"]["email_address"]
    assert message.from_.name == expected["from_"]["name"]

    def check_recipient_list(actual, exp_list):
        assert len(actual) == len(exp_list)
        for rec, exp in zip(actual, exp_list, strict=False):
            assert rec.email_address == exp["email_address"]
            assert rec.name == exp["name"]

    check_recipient_list(message.to_recipients, expected["to_recipients"])
    check_recipient_list(message.cc_recipients, expected["cc_recipients"])
    check_recipient_list(message.bcc_recipients, expected["bcc_recipients"])
    check_recipient_list(message.reply_to, expected["reply_to"])


@pytest.mark.parametrize(
    "initial, add_params, expected_to_recipients",
    [
        # Add a "To" recipient
        (
            {"to_recipients": []},
            {"to_add": ["new@example.com"]},
            [{"email_address": "new@example.com", "name": ""}],
        ),
        # Add a "To" recipient that already exists
        (
            {"to_recipients": [{"email_address": "dup@example.com", "name": ""}]},
            {"to_add": ["dup@example.com"]},
            [
                {"email_address": "dup@example.com", "name": ""},
            ],
        ),
        # Remove a "To" recipient
        (
            {
                "to_recipients": [
                    {"email_address": "a@example.com", "name": "A"},
                    {"email_address": "b@example.com", "name": "B"},
                ]
            },
            {"to_remove": ["a@example.com"]},
            [{"email_address": "b@example.com", "name": "B"}],
        ),
        # Add and remove a "To" recipient
        (
            {"to_recipients": [{"email_address": "c@example.com", "name": "C"}]},
            {"to_add": ["d@example.com", "c@example.com"], "to_remove": ["c@example.com"]},
            [{"email_address": "d@example.com", "name": ""}],
        ),
    ],
)
def test_update_recipient_lists(initial, add_params, expected_to_recipients):
    msg = Message()
    msg.to_recipients = [
        Recipient(email_address=r["email_address"], name=r.get("name", ""))
        for r in initial.get("to_recipients", [])
    ]
    msg.update_recipient_lists(
        to_add=add_params.get("to_add"), to_remove=add_params.get("to_remove")
    )
    result = [r.to_dict() for r in msg.to_recipients]
    assert result == expected_to_recipients, f"Expected {expected_to_recipients}, got {result}"



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_x
from arcade_x.tools.tweets import (
    delete_tweet_by_id,
    lookup_tweet_by_id,
    post_tweet,
    search_recent_tweets_by_keywords,
    search_recent_tweets_by_username,
)
from arcade_x.tools.users import lookup_single_user_by_username

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.7,
    warn_threshold=0.9,
)

catalog = ToolCatalog()
# Register the X tools
catalog.add_module(arcade_x)

search_recent_tweets_by_username_history = [
    {"role": "user", "content": "list 1 tweet from elonmusk"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_kineaPbYCAof3n6qCwnYSKBb",
                "type": "function",
                "function": {
                    "name": "X_SearchRecentTweetsByUsername",
                    "arguments": '{"max_results":1,"username":"elonmusk"}',
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"data":[{"author_id":"44196397","author_name":"Elon Musk","author_username":"elonmusk","edit_history_tweet_ids":["1866572304320466985"],"id":"1866572304320466985","text":"RT @chamath: Meanwhile the State of California is going to spend almost double this ($35B) to build a 171 mile stretch of rail between Merc…","tweet_url":"https://x.com/x/status/1866572304320466985"},{"author_id":"44196397","edit_history_tweet_ids":["1866571568266219998"],"id":"1866571568266219998","text":"This is awesome 🚀🇺🇸 https://twitter.com/cb_doge/status/1866565984502550905","tweet_url":"https://x.com/x/status/1866571568266219998"},{"author_id":"44196397","edit_history_tweet_ids":["1866571416969285954"],"id":"1866571416969285954","text":"@ajtourville @Tesla I’ve always felt that the climate predictions were too pessimistic and bound to backfire. \\n\\nExtreme environmentalists can’t say ridiculous things like the world is doomed in 5 years, because 5 years goes by, the world is ok and they lose credibility. \\n\\nIf we transition to… https://x.com/i/web/status/1866571416969285954","tweet_url":"https://x.com/x/status/1866571416969285954"},{"author_id":"44196397","edit_history_tweet_ids":["1866569957309603946"],"id":"1866569957309603946","text":"@shaunmmaguire Yes, please. This is gone on for too long. Enough.","tweet_url":"https://x.com/x/status/1866569957309603946"},{"author_id":"44196397","edit_history_tweet_ids":["1866569078539948491"],"id":"1866569078539948491","text":"@FatEmperor 😂","tweet_url":"https://x.com/x/status/1866569078539948491"},{"author_id":"44196397","edit_history_tweet_ids":["1866554579925577793"],"id":"1866554579925577793","text":"@cb_doge I’m not buying or building a house anywhere","tweet_url":"https://x.com/x/status/1866554579925577793"},{"author_id":"44196397","edit_history_tweet_ids":["1866536009833361915"],"id":"1866536009833361915","text":"RT @amuse: http://x.com/i/article/1866500805211123713","tweet_url":"https://x.com/x/status/1866536009833361915"},{"author_id":"44196397","edit_history_tweet_ids":["1866535704924483739"],"id":"1866535704924483739","text":"@benshapiro 😂","tweet_url":"https://x.com/x/status/1866535704924483739"},{"author_id":"44196397","edit_history_tweet_ids":["1866535550632550854"],"id":"1866535550632550854","text":"@AutismCapital 😂","tweet_url":"https://x.com/x/status/1866535550632550854"},{"author_id":"44196397","edit_history_tweet_ids":["1866535352024043804"],"id":"1866535352024043804","text":"@JDVance Yes","tweet_url":"https://x.com/x/status/1866535352024043804"}],"includes":{"users":[{"id":"44196397","name":"Elon Musk","username":"elonmusk"}]},"meta":{"newest_id":"1866572304320466985","next_token":"b26v89c19zqg8o3frr3tekall7a7ooom3sctaw30rz62l","oldest_id":"1866535352024043804","result_count":10}}',  # noqa: RUF001
        "tool_call_id": "call_kineaPbYCAof3n6qCwnYSKBb",
        "name": "X_SearchRecentTweetsByUsername",
    },
    {
        "role": "assistant",
        "content": 'Here is a recent tweet from Elon Musk: \n\n"This is awesome 🚀🇺🇸" - [Tweet link](https://x.com/x/status/1866571568266219998)',
    },
]


@tool_eval()
def x_eval_suite() -> EvalSuite:
    """Evaluation suite for X (Twitter) tools."""

    suite = EvalSuite(
        name="X Tools Evaluation Suite",
        system_message=(
            "You are an AI assistant with access to the X (Twitter) tools. Use them to "
            "help answer the user's X-related tasks/questions."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    # Add cases
    suite.add_case(
        name="Post a tweet",
        user_message=(
            "Send out a tweet that says 'Hello World! Exciting stuff is happening over "
            "at Arcade AI!'"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=post_tweet,
                args={"tweet_text": "Hello World! Exciting stuff is happening over at Arcade AI!"},
            )
        ],
        critics=[
            BinaryCritic(
                critic_field="tweet_text",
                weight=1.0,
            ),
        ],
    )

    suite.add_case(
        name="Delete a tweet by ID",
        user_message="Please delete the tweet with ID '148975632'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=delete_tweet_by_id,
                args={"tweet_id": "148975632"},
            )
        ],
        critics=[
            BinaryCritic(
                critic_field="tweet_id",
                weight=1.0,
            ),
        ],
    )

    suite.add_case(
        name="Search recent tweets by username",
        user_message="Show me the recent tweets from 'elonmusk'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_recent_tweets_by_username,
                args={"username": "elonmusk", "max_results": 10},
            )
        ],
        critics=[
            BinaryCritic(
                critic_field="username",
                weight=1.0,
            ),
        ],
    )

    suite.add_case(
        name="Search recent tweets by username with history",
        user_message="Get the next 42",
        additional_messages=search_recent_tweets_by_username_history,
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_recent_tweets_by_username,
                args={
                    "username": "elonmusk",
                    "max_results": 42,
                    "next_token": "b26v89c19zqg8o3frr3tekall7a7ooom3sctaw30rz62l",
                },
            ),
        ],
        critics=[
            BinaryCritic(
                critic_field="username",
                weight=0.2,
            ),
            BinaryCritic(
                critic_field="max_results",
                weight=0.2,
            ),
            BinaryCritic(
                critic_field="next_token",
                weight=0.6,
            ),
        ],
    )

    suite.add_case(
        name="Lookup user by username",
        user_message="Can you get information about the user '@jack'?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=lookup_single_user_by_username,
                args={"username": "jack"},
            ),
        ],
        critics=[
            BinaryCritic(
                critic_field="username",
                weight=1.0,
            ),
        ],
    )

    # Add a case for searching recent tweets by keywords
    suite.add_case(
        name="Search recent tweets by keywords",
        user_message="Find recent tweets containing 'Arcade AI'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_recent_tweets_by_keywords,
                args={
                    "keywords": None,
                    "phrases": ["Arcade AI"],
                    "max_results": 10,
                },
            ),
        ],
        critics=[
            BinaryCritic(
                critic_field="keywords",
                weight=0.1,
            ),
            BinaryCritic(
                critic_field="phrases",
                weight=0.9,
            ),
        ],
    )

    # Extend the case to test lookup_tweet_by_id
    suite.extend_case(
        name="Lookup tweet by ID",
        user_message="Can you provide details about the tweet with ID '123456789'?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=lookup_tweet_by_id,
                args={"tweet_id": "123456789"},
            ),
        ],
        critics=[
            BinaryCritic(
                critic_field="tweet_id",
                weight=1.0,
            ),
        ],
    )

    return suite



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from arcade_tdk.errors import ToolExecutionError

from arcade_search.constants import DEFAULT_GOOGLE_NEWS_COUNTRY, DEFAULT_GOOGLE_NEWS_LANGUAGE
from arcade_search.exceptions import CountryNotFoundError, LanguageNotFoundError
from arcade_search.google_data import COUNTRY_CODES, LANGUAGE_CODES
from arcade_search.utils import call_serpapi, extract_news_results, prepare_params


@tool(requires_secrets=["SERP_API_KEY"])
async def search_news_stories(
    context: ToolContext,
    keywords: Annotated[
        str,
        "Keywords to search for news articles. E.g. 'Apple launches new iPhone'.",
    ],
    country_code: Annotated[
        str | None,
        "2-character country code to search for news articles. E.g. 'us' (United States). "
        f"Defaults to '{DEFAULT_GOOGLE_NEWS_COUNTRY}'.",
    ] = None,
    language_code: Annotated[
        str,
        "2-character language code to search for news articles. E.g. 'en' (English). "
        f"Defaults to '{DEFAULT_GOOGLE_NEWS_LANGUAGE}'.",
    ] = DEFAULT_GOOGLE_NEWS_LANGUAGE,
    limit: Annotated[
        int | None,
        "Maximum number of news articles to return. Defaults to None "
        "(returns all results found by the API).",
    ] = None,
) -> Annotated[dict[str, list[dict[str, Any]]], "News results."]:
    """Search for news articles related to a given query."""
    if not keywords:
        raise ToolExecutionError("Keywords are required to search for news articles.")

    if country_code and country_code not in COUNTRY_CODES:
        raise CountryNotFoundError(country_code)

    if language_code not in LANGUAGE_CODES:
        raise LanguageNotFoundError(language_code)

    params = prepare_params("google_news", q=keywords, gl=country_code, hl=language_code)
    results = call_serpapi(context, params)
    return {"news_results": extract_news_results(results, limit=limit)}



import json

from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_asana
from arcade_asana.tools import list_workspaces

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_asana)


@tool_eval()
def list_workspaces_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="list workspaces eval suite",
        system_message=(
            "You are an AI assistant with access to Asana tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List workspaces",
        user_message="List the workspaces in Asana.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_workspaces,
                args={
                    "limit": 100,
                    "offset": 0,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="offset", weight=0.5),
        ],
    )

    suite.add_case(
        name="List workspaces with pagination",
        user_message="Show me the next 2 workspaces.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_workspaces,
                args={
                    "limit": 2,
                    "offset": 2,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="offset", weight=0.5),
        ],
        additional_messages=[
            {"role": "user", "content": "Show me 2 workspaces in Asana."},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Asana_ListWorkspaces",
                            "arguments": '{"limit":2}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "count": 2,
                    "workspaces": [
                        {
                            "id": "1234567890",
                            "name": "Workspace Hello",
                        },
                        {
                            "id": "1234567891",
                            "name": "Workspace World",
                        },
                    ],
                }),
                "tool_call_id": "call_1",
                "name": "Asana_ListWorkspaces",
            },
            {
                "role": "assistant",
                "content": "Here are two workspaces in Asana:\n\n1. Workspace Hello\n2. Workspace World",
            },
        ],
    )

    suite.add_case(
        name="List workspaces with pagination changing the limit",
        user_message="Show me the next 5 workspaces.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_workspaces,
                args={
                    "limit": 5,
                    "offset": "abc123",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="offset", weight=0.5),
        ],
        additional_messages=[
            {"role": "user", "content": "Show me 5 workspaces in Asana."},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Asana_ListWorkspaces",
                            "arguments": '{"limit":5}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "count": 2,
                    "next_page": {
                        "has_more_results": True,
                        "next_page_token": "abc123",
                    },
                    "workspaces": [
                        {
                            "id": "1234567890",
                            "name": "Workspace Hello",
                        },
                        {
                            "id": "1234567891",
                            "name": "Workspace World",
                        },
                    ],
                }),
                "tool_call_id": "call_1",
                "name": "Asana_ListWorkspaces",
            },
            {
                "role": "assistant",
                "content": "Here are two workspaces in Asana:\n\n1. Workspace Hello\n2. Workspace World",
            },
        ],
    )

    return suite



import decimal
import math
from decimal import Decimal
from typing import Annotated

from arcade_tdk import tool

decimal.getcontext().prec = 100


@tool
def deg_to_rad(
    degrees: Annotated[str, "Angle in degrees as a string"],
) -> Annotated[str, "Angle in radians as a string"]:
    """
    Convert an angle from degrees to radians.
    """
    # Use Decimal for arbitrary precision
    return str(math.radians(Decimal(degrees)))


@tool
def rad_to_deg(
    radians: Annotated[str, "Angle in radians as a string"],
) -> Annotated[str, "Angle in degrees as a string"]:
    """
    Convert an angle from radians to degrees.
    """
    # Use Decimal for arbitrary precision
    return str(math.degrees(Decimal(radians)))



from typing import Annotated

import pytest
from arcade_core.catalog import ToolCatalog, get_wire_type
from arcade_tdk import tool


class Case:
    def __init__(self, input_type: type, output_type: type | None):
        self.input_type = input_type
        self.output_type = output_type

    def __str__(self):
        return f"Case(input_type={self.input_type}, output_type={self.output_type})"


primitives = [bool, float, int, str]

test_cases = [
    Case(input_type=input_type, output_type=output_type)
    for input_type in [*primitives, []]
    for output_type in [*primitives, None]
] + [
    Case(input_type=[primitives[i] for i in range(n)], output_type=output_type)
    for n in range(2, len(primitives) + 1)
    for output_type in [*primitives, None]
]


# Generate tool functions dynamically
def generate_tool_function(input_types: list[type], output_type: type | None):
    input_annotation = ", ".join([
        f"param{i}: Annotated[{input_type.__name__}, 'Param {i + 1}']"
        for i, input_type in enumerate(input_types)
    ])
    output_annotation = f" -> {output_type.__name__}" if output_type else ""

    func_code = f"""
@tool(desc="Generated function with input and output types")
def generated_func({input_annotation}){output_annotation}:
    pass
"""
    local_vars = {}
    exec(func_code, {"tool": tool, "Annotated": Annotated}, local_vars)  # noqa: S102
    generated_func = local_vars.get("generated_func")
    generated_func.__source__ = func_code  # Attach the source code to the function
    return generated_func


@pytest.mark.parametrize("test_case", test_cases, ids=[str(tc) for tc in test_cases])
def test_create_tool_def2(test_case):
    input_types = (
        test_case.input_type if isinstance(test_case.input_type, list) else [test_case.input_type]
    )
    output_type = test_case.output_type

    # Generate the function dynamically
    generated_func = generate_tool_function(input_types, output_type)

    assert generated_func is not None, "generated_func was not created"

    # Create tool definition using the generated function
    tool_def = ToolCatalog.create_tool_definition(generated_func, "1.0")

    for i, input_type in enumerate(input_types):
        param = tool_def.input.parameters[i]
        assert (
            param.value_schema.val_type == get_wire_type(input_type)
        ), f"Parameter {param.name} has value type {param.value_schema.val_type} but {input_type} was expected at index {i}"

    if output_type:
        assert tool_def.output.value_schema.val_type == get_wire_type(
            output_type
        ), f"Output has value type {tool_def.output.val_type} but {output_type} was expected"
    else:
        assert tool_def.output.value_schema is None, "Output is not None"



import pytest
from arcade_tdk import ToolAuthorizationContext, ToolContext


@pytest.fixture
def tool_context():
    """Fixture for the ToolContext with mock authorization."""
    return ToolContext(
        authorization=ToolAuthorizationContext(token="test_token", user_info={"sub": "test_user"}),  # noqa: S106
        user_id="test_user",
    )


@pytest.fixture
def mock_httpx_client(mocker):
    """Fixture to mock the httpx.AsyncClient."""
    # Mock the AsyncClient context manager
    mock_client = mocker.patch("httpx.AsyncClient", autospec=True)
    async_mock_client = mock_client.return_value.__aenter__.return_value
    return async_mock_client
